# -*- coding: utf-8 -*-
"""*A similarity preserving perceptual hash for images.*

The ISCC Content-Code Image is created by calculating a discrete cosine transform on
normalized image-pixels and comparing the values from the upper left area of the
dct-matrix against their median values to set the hash-bits.

Images must be normalized before using gen_image_code. Prepare images as follows:

- Transpose image according to EXIF Orientation
- Add white background to image if it has alpha transparency
- Crop empty borders of image
- Convert image to grayscale
- Resize image to 32x32
- Flatten 32x32 matrix to an array of 1024 grayscale (uint8) pixel values
"""
from statistics import median
from typing import Sequence
from more_itertools import chunked
import iscc_core as ic


__all__ = [
    "gen_image_code",
    "gen_image_code_v0",
    "soft_hash_image_v0",
]


def gen_image_code(pixels, bits=ic.core_opts.image_bits):
    # type: (Sequence[int], int) -> dict
    """
    Create an ISCC Content-Code Image with the latest standard algorithm.

    :param Sequence[int] pixels: Normalized image pixels (32x32 flattened gray values).
    :param int bits: Bit-length of ISCC Content-Code Image (default 64).
    :return: ISCC object with Content-Code Image.
    :rtype: ISCC
    """
    return gen_image_code_v0(pixels, bits)


def gen_image_code_v0(pixels, bits=ic.core_opts.image_bits):
    # type: (Sequence[int], int) -> dict
    """
    Create an ISCC Content-Code Image with algorithm v0.

    :param Sequence[int] pixels: Normalized image pixels (32x32 flattened gray values)
    :param int bits: Bit-length of ISCC Content-Code Image (default 64).
    :return: ISCC object with Content-Code Image.
    :rtype: ISCC
    """
    digest = soft_hash_image_v0(pixels, bits=bits)
    image_code = ic.encode_component(
        mtype=ic.MT.CONTENT,
        stype=ic.ST_CC.IMAGE,
        version=ic.VS.V0,
        bit_length=bits,
        digest=digest,
    )
    iscc = "ISCC:" + image_code
    return {"iscc": iscc}


def soft_hash_image_v0(pixels, bits=ic.core_opts.image_bits):
    # type: (Sequence[int], int) -> bytes
    """
    Calculate image hash from normalized grayscale pixel sequence of length 1024.

    :param Sequence[int] pixels:
    :param int bits: Bit-length of image hash (default 64).
    :return: Similarity preserving Image-Hash digest.
    :rtype: bytes
    """
    if not bits <= 256:
        raise AssertionError(f"{bits} bits exeeds max lenght 256 for soft_hash_image")

    # DCT per row
    dct_row_lists = []
    for pixel_list in chunked(pixels, 32):
        dct_row_lists.append(ic.alg_dct(pixel_list))

    # DCT per col
    dct_row_lists_t = list(map(list, zip(*dct_row_lists)))
    dct_col_lists_t = []
    for dct_list in dct_row_lists_t:
        dct_col_lists_t.append(ic.alg_dct(dct_list))

    dct_matrix = list(map(list, zip(*dct_col_lists_t)))

    def flatten(m, x, y):
        """Extract and flatten an 8 x 8 slice from a 2d matrix starting at col/row."""
        return [v for sublist in m[y : y + 8] for v in sublist[x : x + 8]]

    bitstring = ""
    slices = ((0, 0), (1, 0), (0, 1), (1, 1))

    for xy in slices:
        # Extract 8 x 8 slice
        flat_list = flatten(dct_matrix, *xy)

        # Calculate median
        med = median(flat_list)

        # Append 64-bit digest by comparing to median
        for value in flat_list:
            if value > med:
                bitstring += "1"
            else:
                bitstring += "0"
        bl = len(bitstring)
        if bl >= bits:
            hash_digest = int(bitstring, 2).to_bytes(bl // 8, "big", signed=False)
            return hash_digest



# -*- coding: utf-8 -*-
"""
Build cython extension modules.

The shared library can also be built manually using the command:
$ cythonize -X language_level=3 -a -i ./iscc_core/cdc.py
$ cythonize -X language_level=3 -a -i ./iscc_core/minhash.py
$ cythonize -X language_level=3 -a -i ./iscc_core/simhash.py
"""
try:
    from Cython.Build import cythonize, build_ext
except ImportError:
    # dummy build function for poetry
    def build(setup_kwargs):
        pass

else:

    class build_ext_gracefull(build_ext):
        def run(self):
            try:
                print("Trying to compile C accelerator modules")
                super().run()
                print("Successfully comiled C accelerator modules")
            except Exception as e:
                print(e)
                print("********************************************************************")
                print("Failed to compile C accelerator module, falling back to pure python.")
                print("********************************************************************")

        def build_extensions(self):
            try:
                print("Trying to compile C accelerator modules")
                super().build_extensions()
                print("Successfully comiled C accelerator modules")
            except Exception as e:
                print(e)
                print("********************************************************************")
                print("Failed to compile C accelerator module, falling back to pure python.")
                print("********************************************************************")

    def build(setup_kwargs):
        try:
            setup_kwargs.update(
                dict(
                    ext_modules=cythonize(
                        [
                            "iscc_core/cdc.py",
                            "iscc_core/minhash.py",
                            "iscc_core/simhash.py",
                        ]
                    ),
                    cmdclass=dict(build_ext=build_ext_gracefull),
                )
            )
        except Exception as e:
            print(e)
            print("********************************************************************")
            print("Failed to compile C accelerator module, falling back to pure python.")
            print("********************************************************************")



# -*- coding: utf-8 -*-
"""*A similarity preserving hash for video content*

The **Content-Code Video** is generated from MPEG-7 video frame signatures.
The [iscc-sdk](https://github.com/iscc/iscc-sdk) uses [ffmpeg](https://www.ffmpeg.org) to
extract frame signatures with the following command line parameters:

`$ ffmpeg -i video.mpg -vf fps=fps=5,signature=format=xml:filename=sig.xml -f null -`

The relevant frame signatures can be parsed from the following elements in sig.xml:

`<FrameSignature>0  0  0  1  0  0  1  0  1  1  0  0  1  1 ...</FrameSignature>`

!!! tip
    It is also possible to extract the signatures in a more compact binary format.
    But the format requires a custom binary parser to decode the frame signaturs.
"""
from typing import Sequence, Tuple
import iscc_core as ic


def gen_video_code(frame_sigs, bits=ic.core_opts.video_bits):
    # type: (Sequence[ic.FrameSig], int) -> dict
    """
    Create an ISCC Video-Code with the latest standard algorithm.

    :param ic.FrameSig frame_sigs: Sequence of MP7 frame signatures
    :param int bits: Bit-length resulting Instance-Code (multiple of 64)
    :return: ISCC object with Video-Code
    :rtype: dict
    """
    return gen_video_code_v0(frame_sigs, bits)


def gen_video_code_v0(frame_sigs, bits=ic.core_opts.video_bits):
    # type: (Sequence[ic.FrameSig], int) -> dict
    """
    Create an ISCC Video-Code with algorithm v0.

    :param ic.FrameSig frame_sigs: Sequence of MP7 frame signatures
    :param int bits: Bit-length resulting Video-Code (multiple of 64)
    :return: ISCC object with Video-Code
    :rtype: dict
    """
    digest = soft_hash_video_v0(frame_sigs, bits=bits)
    video_code = ic.encode_component(
        mtype=ic.MT.CONTENT,
        stype=ic.ST_CC.VIDEO,
        version=ic.VS.V0,
        bit_length=bits,
        digest=digest,
    )
    iscc = "ISCC:" + video_code
    return dict(iscc=iscc)


def soft_hash_video_v0(frame_sigs, bits=ic.core_opts.video_bits):
    # type: (Sequence[ic.FrameSig], int) -> bytes
    """
    Compute video hash v0 from MP7 frame signatures.

    :param ic.FrameSig frame_sigs: 2D matrix of MP7 frame signatures
    :param int bits: Bit-length of resulting Video-Code (multiple of 64)
    """

    if not isinstance(frame_sigs[0], tuple):
        frame_sigs = [tuple(sig) for sig in frame_sigs]
    sigs = set(frame_sigs)
    vecsum = [sum(col) for col in zip(*sigs)]
    video_hash_digest = ic.alg_wtahash(vecsum, bits)
    return video_hash_digest



# -*- coding: utf-8 -*-
"""*A similarity perserving hash for binary data (soft hash).*"""
from typing import Optional
import xxhash
import iscc_core as ic

__all__ = [
    "gen_data_code",
    "gen_data_code_v0",
    "soft_hash_data_v0",
    "DataHasher",
    "DataHasherV0",
]


def gen_data_code(stream, bits=ic.core_opts.data_bits):
    # type: (ic.Stream, int) -> dict
    """
    Create a similarity preserving ISCC Data-Code with the latest standard algorithm.

    :param Stream stream: Input data stream.
    :param int bits: Bit-length of ISCC Data-Code (default 64).
    :return: ISCC Data-Code
    :rtype: dict
    """
    return gen_data_code_v0(stream, bits)


def gen_data_code_v0(stream, bits=ic.core_opts.data_bits):
    # type: (ic.Stream, int) -> dict
    """
    Create an ISCC Data-Code with algorithm v0.

    :param Stream stream: Input data stream.
    :param int bits: Bit-length of ISCC Data-Code (default 64).
    :return: ISCC object with Data-Code
    :rtype: dict
    """

    hasher = DataHasherV0()
    data = stream.read(ic.core_opts.io_read_size)

    while data:
        hasher.push(data)
        data = stream.read(ic.core_opts.io_read_size)

    data_code = hasher.code(bits=bits)
    iscc = "ISCC:" + data_code
    return dict(iscc=iscc)


def soft_hash_data_v0(stream):
    # type: (ic.Stream) -> bytes
    """
    Create a similarity preserving Data-Hash digest

    :param Stream stream: Input data stream.
    :return: 256-bit Data-Hash (soft-hash) digest used as body for Data-Code
    :rtype: bytes
    """
    hasher = DataHasherV0()
    data = stream.read(ic.core_opts.io_read_size)

    while data:
        hasher.push(data)
        data = stream.read(ic.core_opts.io_read_size)
    return hasher.digest()


class DataHasherV0:
    """Incremental Data-Hash generator."""

    def __init__(self, data=None):
        # type: (Optional[ic.Data]) -> None
        """
        Create a DataHasher

        :param Optional[Data] data: initial payload for hashing.
        """
        self.chunk_features = []
        self.chunk_sizes = []
        self.tail = None
        data = data or b""
        self.push(data)

    def push(self, data):
        # type: (ic.Data) -> None
        """Push data to the Data-Hash generator."""
        if self.tail:
            data = self.tail + data

        for chunk in ic.alg_cdc_chunks(
            data, utf32=False, avg_chunk_size=ic.core_opts.data_avg_chunk_size
        ):
            self.chunk_sizes.append(len(chunk))
            self.chunk_features.append(xxhash.xxh32_intdigest(chunk))
            self.tail = chunk  # Last chunk may not be final

        self.chunk_features = self.chunk_features[:-1]
        self.chunk_sizes = self.chunk_sizes[:-1]

    def digest(self):
        # type: () -> bytes
        """Calculate 256-bit minhash digest from feature hashes."""
        self._finalize()
        return ic.alg_minhash_256(self.chunk_features)

    def code(self, bits=ic.core_opts.data_bits):
        # type: (int) -> str
        """
        Encode digest as an ISCC Data-Code unit.

        :param int bits: Number of bits for the ISCC Data-Code
        :return: ISCC Data-Code
        :rtype: str
        """
        data_code = ic.encode_component(
            mtype=ic.MT.DATA,
            stype=ic.ST.NONE,
            version=ic.VS.V0,
            bit_length=bits,
            digest=self.digest(),
        )
        return data_code

    def _finalize(self):
        if self.tail is not None:
            self.chunk_features.append(xxhash.xxh32_intdigest(self.tail))
            self.chunk_sizes.append(len(self.tail))
            self.tail = None


DataHasher = DataHasherV0



# -*- coding: utf-8 -*-
"""*A data checksum.*"""
from blake3 import blake3
from typing import Optional
import iscc_core as ic


__all__ = [
    "gen_instance_code",
    "gen_instance_code_v0",
    "hash_instance_v0",
    "InstanceHasher",
    "InstanceHasherV0",
]


def gen_instance_code(stream, bits=ic.core_opts.instance_bits):
    # type: (ic.Stream, int) -> dict
    """
    Create an ISCC Instance-Code with the latest standard algorithm.

    :param Stream stream: Binary data stream for Instance-Code generation
    :param int bits: Bit-length resulting Instance-Code (multiple of 64)
    :return: ISCC object with properties: iscc, datahash, filesize
    :rtype: dict
    """
    return gen_instance_code_v0(stream, bits)


def gen_instance_code_v0(stream, bits=ic.core_opts.instance_bits):
    # type: (ic.Stream, int) -> dict
    """
    Create an ISCC Instance-Code with algorithm v0.

    :param Stream stream: Binary data stream for Instance-Code generation
    :param int bits: Bit-length of resulting Instance-Code (multiple of 64)
    :return: ISCC object with Instance-Code and properties: datahash, filesize
    :rtype: dict
    """
    hasher = InstanceHasherV0()
    data = stream.read(ic.core_opts.io_read_size)
    while data:
        hasher.push(data)
        data = stream.read(ic.core_opts.io_read_size)

    instance_code = hasher.code(bits=bits)
    iscc = "ISCC:" + instance_code
    instance_code_obj = dict(
        iscc=iscc,
        datahash=hasher.multihash(),
        filesize=hasher.filesize,
    )

    return instance_code_obj


def hash_instance_v0(stream):
    # type: (ic.Stream) -> bytes
    """
    Create 256-bit hash digest for the Instance-Code body

    :param Stream stream: Binary data stream for hash generation.
    :return: 256-bit Instance-Hash digest used as body of Instance-Code
    :rtype: bytes
    """
    hasher = InstanceHasherV0()
    data = stream.read(ic.core_opts.io_read_size)
    while data:
        hasher.push(data)
        data = stream.read(ic.core_opts.io_read_size)
    return hasher.digest()


class InstanceHasherV0:
    """Incremental Instance-Hash generator."""

    #: Multihash prefix
    mh_prefix: bytes = b"\x1e\x20"

    def __init__(self, data=None):
        # type: (Optional[ic.Data]) -> None
        self.hasher = blake3(max_threads=blake3.AUTO)
        self.filesize = 0
        data = data or b""
        self.push(data)

    def push(self, data):
        # type: (ic.Data) -> None
        """
        Push data to the Instance-Hash generator.

        :param Data data: Data to be hashed
        """
        self.filesize += len(data)
        self.hasher.update(data)

    def digest(self):
        # type: () -> bytes
        """
        Return Instance-Hash

        :return: Instance-Hash digest
        :rtype: bytes
        """
        return self.hasher.digest()

    def multihash(self):
        # type: () -> str
        """
        Return blake3 multihash

        :return: Blake3 hash as 256-bit multihash
        :rtype: str
        """
        return (self.mh_prefix + self.digest()).hex()

    def code(self, bits=ic.core_opts.instance_bits):
        # type: (int) -> str
        """
        Encode digest as an ISCC Instance-Code unit.

        :param int bits: Number of bits for the ISCC Instance-Code
        :return: ISCC Instance-Code
        :rtype: str
        """
        instance_code = ic.encode_component(
            mtype=ic.MT.INSTANCE,
            stype=ic.ST.NONE,
            version=ic.VS.V0,
            bit_length=bits,
            digest=self.digest(),
        )
        return instance_code


InstanceHasher = InstanceHasherV0



# -*- coding: utf-8 -*-
"""*A unique, time-sorted identifier composed of an 48-bit timestamp and 16 to 208 bit randomness.*

The ISCC Flake-Code is a unique identifier for distributed ID generation. The 64-bit version
can be used as efficient surrogate key in database systems. It has guaranteed uniqueness if
generated from a singele process and is time sortable in integer and base32hex representation.
The 128-bit version is a K-sortable, globally unique identifier for use in distributed systems and
is compatible with UUID.

!!! example

    ```python
    >>> import iscc_core as ic
    >>> ic.gen_flake_code(bits=64)
    {'iscc': 'ISCC:OAAQC7YN7PG2XOR4'}

    >>> ic.gen_flake_code(bits=128)
    {'iscc': 'ISCC:OABQC7YN7RJGUUTLKDSKBXO25MA5E'}

    # Or use the convenience Flake class for easy access to different representations

    >>> flake = ic.Flake(bits=64)
    >>> flake.iscc
    'ISCC:OAAQC7YOADBZYNF7'

    >>> flake.time
    '2022-02-18T18:03:25.468'

    >>> flake.int
    107820312524764351

    >>> flake.string
    '05VGS063JGQBU'
    ```
"""
import os
import time
from collections import Counter
from typing import Optional
import iscc_core as ic


__all__ = [
    "gen_flake_code",
    "gen_flake_code_v0",
    "uid_flake_v0",
]

_COUNTER = Counter()


def gen_flake_code(bits=ic.core_opts.flake_bits):
    # type: (int) -> dict
    """
    Create an ISCC Flake-Code with the latest standard algorithm

    :param int bits: Target bit-length of generated Flake-Code
    :return: ISCC object with Flake-Code
    :rtype: dict
    """
    return gen_flake_code_v0(bits=bits)


def gen_flake_code_v0(bits=ic.core_opts.flake_bits):
    """
    Create an ISCC Flake-Code with the latest algorithm v0

    :param int bits: Target bit-length of generated Flake-Code
    :return: ISCC object with Flake-Code
    :rtype: dict
    """

    digest = uid_flake_v0(bits=bits)
    flake_code = ic.encode_component(
        mtype=ic.MT.FLAKE, stype=ic.ST.NONE, version=ic.VS.V0, bit_length=bits, digest=digest
    )
    iscc = "ISCC:" + flake_code
    return {"iscc": iscc}


def uid_flake_v0(ts=None, bits=ic.core_opts.flake_bits):
    # type: (Optional[float], int) -> bytes
    """
    Generate time and randomness based Flake-Hash

    :param Optional[float] ts: Unix timestamp (defaults to current time)
    :param int bits: Bit-length resulting Flake-Code (multiple of 32)
    :return: Flake-Hash digest
    :rtype: bytes
    """
    if not 64 <= bits <= 256:
        raise ValueError(f"{bits} bits for flake outside 64 - 256 bits")
    if bits % 32:
        raise ValueError(f"{bits} bits for flake is not divisible by 32")

    nbytes_rnd = (bits // 8) - 6

    ts = time.time() if ts is None else ts
    milliseconds = int(ts * 1000)
    timedata = milliseconds.to_bytes(6, "big", signed=False)
    counter = _COUNTER[timedata]
    if counter == 0:
        # Init counter with random value
        counterdata = os.urandom(nbytes_rnd)
        startcount = int.from_bytes(counterdata, "big", signed=False)
        _COUNTER[timedata] = startcount
        counter = startcount
    counterdata = counter.to_bytes(nbytes_rnd, "big", signed=False)
    _COUNTER.update([timedata])
    flake_digest = timedata + counterdata
    return flake_digest



import unicodedata
import jcs
from more_itertools import interleave, sliced
from blake3 import blake3
from typing import Optional, Union
from data_url import DataURL
import iscc_core as ic


__all__ = [
    "gen_meta_code",
    "gen_meta_code_v0",
    "soft_hash_meta_v0",
    "text_trim",
    "text_remove_newlines",
    "text_clean",
]


def gen_meta_code(name, description=None, meta=None, bits=ic.core_opts.meta_bits):
    # type: (str, Optional[str], Optional[ic.Meta], int) -> dict
    """
    Create an ISCC Meta-Code using the latest standard algorithm.

    :param str name: Name or title of the work manifested by the digital asset
    :param Optional[str] description: Optional description for disambiguation
    :param Optional[Union[dict,str] meta: Dict or Data-URL string with extended metadata
    :param int bits: Bit-length of resulting Meta-Code (multiple of 64)
    :return: ISCC object with Meta-Code and properties name, description, properties, metahash
    :rtype: dict
    """
    return gen_meta_code_v0(name, description=description, meta=meta, bits=bits)


def gen_meta_code_v0(name, description=None, meta=None, bits=ic.core_opts.meta_bits):
    # type: (str, Optional[str], Optional[ic.Meta], int) -> dict
    """
    Create an ISCC Meta-Code with the algorithm version 0.

    :param str name: Name or title of the work manifested by the digital asset
    :param Optional[str] description: Optional description for disambiguation
    :param Optional[Union[dict,str] meta: Dict or Data-URL string with extended metadata
    :param int bits: Bit-length of resulting Meta-Code (multiple of 64)
    :return: ISCC object with possible fields: iscc, name, description, metadata, metahash
    :rtype: dict
    """

    # 1. Normalize `name`
    name = "" if name is None else name
    name = text_clean(name)
    name = text_remove_newlines(name)
    name = text_trim(name, ic.core_opts.meta_trim_name)

    if not name:
        raise ValueError("Meta-Code requires non-empty name element (after normalization)")

    # 2. Normalize `description`
    description = "" if description is None else description
    description = text_clean(description)
    description = text_trim(description, ic.core_opts.meta_trim_description)

    # Calculate meta_code, metahash, and output metadata values for the different input cases
    if meta:
        if isinstance(meta, str):
            # Data-URL expected
            durl = meta
            payload = DataURL.from_url(durl).data
            meta_code_digest = soft_hash_meta_v0(name, payload)
            metahash = ic.multi_hash_blake3(payload)
            metadata_value = durl
        elif isinstance(meta, dict):
            payload = jcs.canonicalize(meta)
            meta_code_digest = soft_hash_meta_v0(name, payload)
            metahash = ic.multi_hash_blake3(payload)
            media_type = "application/ld+json" if "@context" in meta else "application/json"
            durl_obj = DataURL.from_data(media_type, base64_encode=True, data=payload)
            metadata_value = durl_obj.url
        else:
            raise TypeError(f"metadata must be Data-URL string or dict not {type(meta)}")
    else:
        payload = " ".join((name, description)).strip().encode("utf-8")
        meta_code_digest = soft_hash_meta_v0(name, description)
        metahash = ic.multi_hash_blake3(payload)
        metadata_value = None

    meta_code = ic.encode_component(
        mtype=ic.MT.META,
        stype=ic.ST.NONE,
        version=ic.VS.V0,
        bit_length=bits,
        digest=meta_code_digest,
    )
    iscc = "ISCC:" + meta_code

    # Build result
    result = {"iscc": iscc}
    if name:
        result["name"] = name
    if description:
        result["description"] = description
    if metadata_value:
        result["meta"] = metadata_value

    result["metahash"] = metahash

    return result


def soft_hash_meta_v0(name, extra=None):
    # type: (str, Union[str,bytes,None]) -> bytes
    """
    Calculate simmilarity preserving 256-bit hash digest from asset metadata.

    Textual input should be stripped of markup, normalized and trimmed before hashing.
    Bytes input can be any serialized metadata (JSON, XML, Image...). Metadata should be
    serialized in a canonical form (for example
    [JCS](https://tools.ietf.org/id/draft-rundgren-json-canonicalization-scheme-00.html) for JSON)

    !!! note
        The processing algorithm depends on the type of the `extra` input.
        If the `extra` field is supplied and non-empty, we create separate hashes for
        `name` and `extra` and interleave them in 32-bit chunks:

        - If the `extra` input is `None` or an empty `str`/`bytes` object the Meta-Hash will
        be generated from the `name`-field only.

        - If the `extra`-input is a non-empty **text** string (str) the string is
        lower-cased and the processing units are utf-8 endoded characters (possibly multibyte).

        - If the `extra`-input is a non-empty **bytes** object the processing is done bytewise.


    :param str name: Title of the work manifested in the digital asset
    :param Union[str,bytes,None] extra: Additional metadata for disambiguation
    :return: 256-bit simhash digest for Meta-Code
    :rtype: bytes
    """
    name = ic.text_collapse(name)
    name_n_grams = ic.sliding_window(name, width=ic.core_opts.meta_ngram_size_text)
    name_hash_digests = [blake3(s.encode("utf-8")).digest() for s in name_n_grams]
    simhash_digest = ic.alg_simhash(name_hash_digests)

    if extra in {None, "", b""}:
        return simhash_digest
    else:
        # Augment with interleaved hash for extra metadata
        if isinstance(extra, bytes):
            # Raw bytes are handled per byte
            extra_n_grams = ic.sliding_window(extra, width=ic.core_opts.meta_ngram_size_bytes)
            extra_hash_digests = [blake3(ngram).digest() for ngram in extra_n_grams]
        elif isinstance(extra, str):
            # Text is collapsed and handled per character (multibyte)
            extra = ic.text_collapse(extra)
            extra_n_grams = ic.sliding_window(extra, width=ic.core_opts.meta_ngram_size_text)
            extra_hash_digests = [blake3(s.encode("utf-8")).digest() for s in extra_n_grams]
        else:
            raise ValueError("parameter `extra` must be of type str or bytes!")

        extra_simhash_digest = ic.alg_simhash(extra_hash_digests)

        # Interleave first half of name and extra simhashes in 32-bit chunks
        chunks_simhash_digest = sliced(simhash_digest[:16], 4)
        chunks_extra_simhash_digest = sliced(extra_simhash_digest[:16], 4)
        interleaved = interleave(chunks_simhash_digest, chunks_extra_simhash_digest)
        simhash_digest = bytearray()
        for chunk in interleaved:
            simhash_digest += chunk

        simhash_digest = bytes(simhash_digest)

        return simhash_digest


def text_trim(text, nbytes):
    # type: (str, int) -> str
    """Trim text such that its utf-8 encoded size does not exceed `nbytes`."""
    return text.encode("utf-8")[:nbytes].decode("utf-8", "ignore").strip()


def text_remove_newlines(text):
    # type: (str) -> str
    """
    Remove newlines.

    The `name` field serves as a displayable title. We remove newlines and leading and trailing
    whitespace. We also collapse consecutive spaces to single spaces.

    :param text: Text for newline removal
    :return: Single line of text
    :rtype: str
    """
    return " ".join(text.split())


def text_clean(text):
    # type: (str) -> str
    """
    Clean text for display.

    - Normalize with NFKC normalization.
    - Remove Control Characters (except newlines)
    - Reduce multiple consecutive newlines to a maximum of two newlines
    - Strip leading and trailing whitespace
    """

    # Unicode normalize
    text = unicodedata.normalize("NFKC", text)

    # Remove control characters
    text = "".join(
        ch for ch in text if unicodedata.category(ch)[0] != "C" or ch in ic.core_opts.text_newlines
    )

    # Collapse more than two consecutive newlines
    chars = []
    newline_count = 0
    for c in text:
        if c in ic.core_opts.text_newlines:
            if newline_count < 2:
                chars.append("\n")
                newline_count += 1
            continue
        else:
            newline_count = 0
        chars.append(c)
    text = "".join(chars)

    return text.strip()



# -*- coding: utf-8 -*-
from datetime import datetime
import random
from typing import List, Union
import base58
import uvarint
from bitarray import bitarray, frozenbitarray
from bitarray.util import ba2hex, ba2int, count_xor

from iscc_core import core_opts
from iscc_core.constants import IsccAny, UNITS, LN, MT, ST, ST_CC, ST_ID, ST_ISCC, VS, MC_PREFIX
from iscc_core.code_flake import uid_flake_v0

from iscc_core.codec import (
    iscc_clean,
    decode_base32,
    decode_length,
    decode_units,
    encode_base32,
    encode_base64,
    encode_length,
    encode_units,
    decode_header,
    encode_header,
    decode_base32hex,
    encode_base32hex,
    encode_component,
)

__all__ = [
    "Code",
    "Flake",
]


class Code:
    """
    Convenience class to handle different representations of an ISCC.
    """

    def __init__(self, code):
        # type: (IsccAny) -> None
        """
        Initialize a Code object from any kind of representation of an ISCC.

        :param AnyISCC code: Any valid representation of an ISCC
        """
        self._head = None
        self._body = None

        if isinstance(code, Code):
            code_fields = code._head + (code.hash_bytes,)
        elif isinstance(code, str):
            code = iscc_clean(code)
            code_fields = decode_header(decode_base32(code))
        elif isinstance(code, tuple):
            code_fields = code
        elif isinstance(code, bytes):
            code_fields = decode_header(code)
        else:
            raise ValueError(f"Code must be str, bytes, tuple or Code not {type(code)}")

        self._head = code_fields[:-1]
        body = bitarray()
        body.frombytes(code_fields[-1])
        self._body = frozenbitarray(body)

    def __str__(self):
        return self.code

    def __repr__(self):
        return f'Code("{self.code}")'

    def __bytes__(self):
        return self.bytes

    def __iter__(self):
        for f in self._head:
            yield f
        yield self.hash_bytes

    @property
    def code(self) -> str:
        """Standard base32 representation of an ISCC code."""
        return encode_base32(self.bytes)

    @property
    def uri(self) -> str:
        """Standard uri representation of an ISCC code."""
        return f"ISCC:{self.code.upper()}"

    @property
    def bytes(self) -> bytes:
        """Raw bytes of code (including header)."""
        return self.header_bytes + self._body.tobytes()

    @property
    def hex(self) -> str:
        """Hex representation of code (including header)."""
        return self.bytes.hex()

    @property
    def base32hex(self) -> str:
        """Base32hex representation of code (including header)"""
        return encode_base32hex(self.bytes)

    @property
    def uint(self) -> int:
        """Integer representation of code (including header)"""
        return int.from_bytes(self.bytes, "big", signed=False)

    @property
    def type_id(self) -> str:
        """A unique composite type-id (use as name to index codes seperately)."""
        if self.maintype == MT.ISCC:
            mtypes = decode_units(self._head[3])
            length = "".join([t.name[0] for t in mtypes]) + "DI"
        else:
            length = self.length
        return f"{self.maintype.name}-" f"{self.subtype.name}-" f"{self.version.name}-" f"{length}"

    @property
    def explain(self) -> str:
        """Human readble representation of code header."""
        if self.maintype == MT.ID:
            counter_bytes = self.hash_bytes[8:]
            if counter_bytes:
                counter = uvarint.decode(counter_bytes)
                return f"{self.type_id}-{self.hash_bytes[:8].hex()}-{counter.integer}"
        return f"{self.type_id}-{self.hash_hex}"

    @property
    def hash_bytes(self) -> bytes:
        """Byte representation of code (without header)"""
        return self._body.tobytes()

    @property
    def hash_hex(self) -> str:
        """Hex string representation of code (without header)."""
        return ba2hex(self._body)

    @property
    def hash_base32(self) -> str:
        """Base32 representation of code (without header)"""
        return encode_base32(self.hash_bytes)

    @property
    def hash_base32hex(self) -> str:
        """Base32hex representation of code (without header)"""
        return encode_base32hex(self.hash_bytes)

    @property
    def hash_bits(self) -> str:
        """String of 0,1 representing the bits of the code (without header)."""
        return self._body.to01()

    @property
    def hash_ints(self) -> List[int]:
        """List of 0,1 integers representing the bits of the code (without header)."""
        return self._body.tolist()

    @property
    def hash_uint(self) -> int:
        """Unsinged integer representation of the code (without header)."""
        return ba2int(self._body, signed=False)

    @property
    def hash_ba(self) -> frozenbitarray:
        """Bitarray object of the code (without header)."""
        return self._body

    @property
    def header_bytes(self) -> bytes:
        """Byte representation of header prefix of the code"""
        return encode_header(*self._head)

    @property
    def maintype(self) -> MT:
        """Enum maintype of code."""
        return MT(self._head[0])

    @property
    def subtype(self) -> Union[ST, ST_CC, ST_ISCC, ST_ID]:
        """Enum subtype of code."""
        if self.maintype in (MT.CONTENT, MT.SEMANTIC):
            return ST_CC(self._head[1])
        elif self.maintype == MT.ISCC:
            return ST_ISCC(self._head[1])
        elif self.maintype == MT.ID:
            return ST_ID(self._head[1])
        return ST(self._head[1])

    @property
    def version(self) -> VS:
        """Enum version of code."""
        return VS(self._head[2])

    @property
    def length(self) -> int:
        """Length of code hash in number of bits (without header)."""
        return decode_length(self._head[0], self._head[3])

    rgen = random.Random(0)  # nosec

    @classmethod
    def rnd(cls, mt=None, st=None, bits=64, data=None):
        """Returns a syntactically correct random code."""

        # MainType
        mt = cls.rgen.choice(list(MT)) if mt is None else mt

        # SubType
        if st is None:
            if mt == MT.CONTENT:
                st = cls.rgen.choice(list(ST_CC))
            elif mt == MT.ISCC:
                if bits == 320:
                    units = (MT.META, MT.SEMANTIC, MT.CONTENT)
                elif bits == 256:
                    units = (MT.META, cls.rgen.choice((MT.CONTENT, MT.SEMANTIC)))
                elif bits == 192:
                    units = (cls.rgen.choice((MT.META, MT.SEMANTIC, MT.CONTENT)),)
                elif bits == 128:
                    units = tuple()
                else:
                    raise ValueError(f"Unsupported bit length {bits} for random ISCC-CODE.")

                if bits == 128:
                    st = ST_ISCC.SUM
                else:
                    st = cls.rgen.choice((0, 1, 2, 3, 4))
                    # st = st if (MT.SEMANTIC in units or MT.CONTENT in units) else ST_ISCC.SUM
                # bits = len(units) * 64 + 128
            elif mt == MT.ID:
                st = cls.rgen.choice(list(ST_ID))
            else:
                st = cls.rgen.choice(list(ST))
        else:
            if mt == MT.ISCC:
                raise ValueError(f"Custom subtype selection not supported for MT.ISCC")

        # Version
        vs = VS.V0

        # Length
        ln_bits = bits or cls.rgen.choice(list(LN)).value
        if mt == MT.ISCC:
            # TODO fix ramdom ISCC with custom SubType generation
            ln_code = encode_units(units)
        else:
            ln_code = encode_length(mt, bits)

        # Body
        data = (
            cls.rgen.getrandbits(ln_bits).to_bytes(length=ln_bits // 8, byteorder="big")
            if data is None
            else data
        )

        return cls((mt, st, vs, ln_code, data))

    @property
    def mc_bytes(self):
        """ISCC header + body with multicodec prefix."""
        return MC_PREFIX + self.bytes

    @property
    def mf_base16(self) -> str:
        """Multiformats base16 encoded."""
        return "f" + self.mc_bytes.hex()

    @property
    def mf_base32(self) -> str:
        """Multiformats base32 encoded."""
        return "b" + encode_base32(self.mc_bytes).lower()

    @property
    def mf_base32hex(self) -> str:
        """Multiformats base32 encoded."""
        return "v" + encode_base32hex(self.mc_bytes).lower()

    @property
    def mf_base58btc(self) -> str:
        """Multiformats base58btc encoded."""
        return "z" + base58.b58encode(self.mc_bytes).decode("ascii")

    @property
    def mf_base64url(self) -> str:
        """Multiformats base64url encoded."""
        return "u" + encode_base64(self.mc_bytes)

    # TODO: bech32m support
    # @property
    # def bech32m(self):
    #     """Encode as bech32m with hrp `iscc`"""
    #     data = [bech32.CHARSET.find(c) for c in self.code.lower()]
    #     return bech32.bech32_encode(
    #         "iscc", data, bech32.Encoding.BECH32M
    #     )

    def __xor__(self, other) -> int:
        """Use XOR operator for hamming distance calculation."""
        return count_xor(self._body, other._body)

    def __eq__(self, other):
        # type: (Code) -> bool
        return self.code == other.code

    def __hash__(self):
        return self.uint


class Flake:
    """Unique lexicographically k-sortable identifier"""

    def __init__(self, ts=None, bits=core_opts.flake_bits):
        self._flake = uid_flake_v0(ts, bits=bits)
        self._bits = bits

    def __repr__(self):
        return f'Flake("{self}")'

    def __str__(self):
        return encode_base32hex(self._flake)

    def __int__(self):
        return int.from_bytes(self._flake, "big", signed=False)

    def __eq__(self, other):
        return self._flake == other._flake

    def __lt__(self, other):
        return self.int < other.int

    def __gt__(self, other):
        return self.int > other.int

    def __hash__(self):
        return self.int

    @property
    def iscc(self):
        """ISCC Flake-Code in canonical ISCC string representation (self-descriptive)"""
        return "ISCC:" + encode_component(
            mtype=MT.FLAKE,
            stype=ST.NONE,
            version=VS.V0,
            bit_length=self._bits,
            digest=self._flake,
        )

    @property
    def time(self):
        """Time component as string in ISO 8601 format"""
        ts = int.from_bytes(self._flake[0:6], "big", signed=False) / 1000
        return datetime.utcfromtimestamp(ts).isoformat(timespec="milliseconds")

    @property
    def int(self):
        """Flake-Code as time-sortable integer (non self-descriptive)"""
        return int(self)

    @property
    def string(self):
        """Flake-Code as short time-sortable base32hex string (non self-descriptive)"""
        return str(self)

    @classmethod
    def from_int(cls, num):
        # type: (int) -> Flake
        """Instantiate  flake from integer."""
        flake_obj = cls()
        flake_obj._flake = num.to_bytes(8, "big", signed=False)
        flake_obj._bits = 64
        return flake_obj

    @classmethod
    def from_string(cls, code):
        """Instantiate flake from string."""
        raw = decode_base32hex(code)
        flake_obj = cls()
        flake_obj._flake = raw
        flake_obj._bits = 64
        return flake_obj



# -*- coding: utf-8 -*-
"""Compatible with [fastcdc](https://pypi.org/project/fastcdc/)"""
import io
from math import log2
from typing import Generator
import iscc_core as ic


__all__ = [
    "alg_cdc_chunks",
]


def alg_cdc_chunks(data, utf32, avg_chunk_size=ic.core_opts.data_avg_chunk_size):
    # type: (Data, bool, int) -> Generator[bytes, None, None]
    """
    A generator that yields data-dependent chunks for `data`.

    Usage Example:

    ```python
    for chunk in cdc_data_chunks(data):
        hash(chunk)
    ```

    :param bytes data: Raw data for variable sized chunking.
    :param bool utf32: If true assume we are chunking text that is utf32 encoded.
    :param int avg_chunk_size: Target chunk size in number of bytes.
    :return: A generator that yields data chunks of variable sizes.
    :rtype: Generator[bytes]
    """

    stream = io.BytesIO(data)
    buffer = stream.read(ic.core_opts.io_read_size)
    if not buffer:
        yield b""

    mi, ma, cs, mask_s, mask_l = alg_cdc_params(avg_chunk_size)

    buffer = memoryview(buffer)
    while buffer:
        if len(buffer) <= ma:
            buffer = memoryview(bytes(buffer) + stream.read(ic.core_opts.io_read_size))
        cut_point = alg_cdc_offset(buffer, mi, ma, cs, mask_s, mask_l)

        # Make sure cut points are at 4-byte aligned for utf32 encoded text
        if utf32:
            cut_point -= cut_point % 4

        yield bytes(buffer[:cut_point])
        buffer = buffer[cut_point:]


def alg_cdc_offset(buffer, mi, ma, cs, mask_s, mask_l):
    # type: (ic.Data, int, int, int, int, int) -> int
    """
    Find breakpoint offset for a given buffer.

    :param Data buffer: The data to be chunked.
    :param int mi: Minimum chunk size.
    :param int ma: Maximung chunk size.
    :param int cs: Center size.
    :param int mask_s: Small mask.
    :param int mask_l: Large mask.
    :return: Offset of dynamic cutpoint in number of bytes.
    :rtype: int
    """

    pattern = 0
    i = mi
    size = len(buffer)
    barrier = min(cs, size)
    while i < barrier:
        pattern = (pattern >> 1) + ic.core_opts.cdc_gear[buffer[i]]
        if not pattern & mask_s:
            return i + 1
        i += 1
    barrier = min(ma, size)
    while i < barrier:
        pattern = (pattern >> 1) + ic.core_opts.cdc_gear[buffer[i]]
        if not pattern & mask_l:
            return i + 1
        i += 1
    return i


def alg_cdc_params(avg_size: int) -> tuple:
    """
    Calculate CDC parameters

    :param int avg_size: Target average size of chunks in number of bytes.
    :returns: Tuple of (min_size, max_size, center_size, mask_s, mask_l).
    """
    ceil_div = lambda x, y: (x + y - 1) // y
    mask = lambda b: 2**b - 1
    min_size = avg_size // 4
    max_size = avg_size * 8
    offset = min_size + ceil_div(min_size, 2)
    center_size = avg_size - offset
    bits = round(log2(avg_size))
    mask_s = mask(bits + 1)
    mask_l = mask(bits - 1)
    return min_size, max_size, center_size, mask_s, mask_l



# -*- coding: utf-8 -*-
"""*A similarity preserving hash for audio content (soft hash).*

Creates an ISCC object that provides an `iscc`-field with an Audio-Code and a
`duration`-field.

The Content-Code Audio is generated from a Chromaprint fingerprint provided as a vector of 32-bit
signed integers. The [iscc-sdk](https://github.com/iscc/iscc-sdk) uses
[fpcalc](https://acoustid.org/chromaprint) to extract Chromaprint vectors with the following
command line parameters:

`$ fpcalc -raw -json -signed -length 0 myaudiofile.mp3`
"""
from typing import Iterable
from more_itertools import divide
import iscc_core as ic


__all__ = [
    "gen_audio_code",
    "gen_audio_code_v0",
    "soft_hash_audio_v0",
]


def gen_audio_code(cv, bits=ic.core_opts.audio_bits):
    # type: (Iterable[int], int) -> dict
    """
    Create an ISCC Content-Code Audio with the latest standard algorithm.

    :param Iterable[int] cv: Chromaprint vector
    :param int bits: Bit-length resulting Content-Code Audio (multiple of 64)
    :return: ISCC object with Content-Code Audio
    :rtype: dict
    """
    return gen_audio_code_v0(cv, bits)


def gen_audio_code_v0(cv, bits=ic.core_opts.audio_bits):
    # type: (Iterable[int], int) -> dict
    """
    Create an ISCC Content-Code Audio with algorithm v0.

    :param Iterable[int] cv: Chromaprint vector
    :param int bits: Bit-length resulting Content-Code Audio (multiple of 64)
    :return: ISCC object with Content-Code Audio
    :rtype: dict
    """
    digest = soft_hash_audio_v0(cv, bits=bits)
    audio_code = ic.encode_component(
        mtype=ic.MT.CONTENT,
        stype=ic.ST_CC.AUDIO,
        version=ic.VS.V0,
        bit_length=bits,
        digest=digest,
    )
    iscc = "ISCC:" + audio_code
    return {"iscc": iscc}


def soft_hash_audio_v0(cv, bits=ic.core_opts.audio_bits):
    # type: (Iterable[int], int) -> bytes
    """
    Create audio similarity hash from a chromaprint vector.

    :param Iterable[int] cv: Chromaprint vector
    :param int bits: Bit-length resulting similarity hash (multiple of 32)
    :return: Audio-Hash digest
    :rtype: bytes
    """

    # Convert chrompaprint vector into list of 4 byte digests
    digests = [int_feature.to_bytes(4, "big", signed=True) for int_feature in cv]

    # Return identity hash if we have 0 digests
    if not digests:
        return b"\x00" * 32

    # Calculate simhash of digests as first 32-bit chunk of the hash
    parts = [ic.alg_simhash(digests)]

    bit_length = 32

    # Calculate separate 32-bit simhashes for each quarter of features (original order)
    for bucket in divide(4, digests):
        features = list(bucket)
        if features:
            parts.append(ic.alg_simhash(features))
        else:
            parts.append(b"\x00\x00\x00\x00")
        bit_length += 32
        if bit_length == bits:
            return b"".join(parts)

    # Calculate separate simhashes for each third of features (ordered by int value)
    cvs = sorted(cv)
    digests = [int_feature.to_bytes(4, "big", signed=True) for int_feature in cvs]
    for bucket in divide(3, digests):
        features = list(bucket)
        if features:
            parts.append(ic.alg_simhash(features))
        else:
            parts.append(b"\x00\x00\x00\x00")
        bit_length += 32
        if bit_length == bits:
            return b"".join(parts)

    return b"".join(parts)



# -*- coding: utf-8 -*-
import math
import re
import uvarint
from typing import List, Tuple
import base58
from bitarray import bitarray
from bitarray.util import int2ba, ba2int
from base64 import b32encode, b32decode
from pybase64 import urlsafe_b64encode, urlsafe_b64decode
from bases import base32hex
from iscc_core.constants import *


base32hexnopad = base32hex.nopad()


########################################################################################
# Core codec functions                                                                 #
########################################################################################


def encode_component(mtype, stype, version, bit_length, digest):
    # type: (MainType, SubType, Version, Length, bytes) -> str
    """
    Encode an ISCC unit inlcuding header and body with standard base32 encoding.

    !!! note
        The `length` value must be the **length in number of bits** for the component.
        If `digest` has more bits than specified by `length` it wil be truncated.


    :param MainType mtype: Maintype of unit (0-6)
    :param SubType stype: SubType of unit depending on MainType (0-5)
    :param Version version: Version of unit algorithm (0).
    :param length bit_length: Length of unit, in number of bits (multiple of 32)
    :param bytes digest: The hash digest of the unit.
    :return: Base32 encoded ISCC-UNIT.
    :rtype: str
    """
    if mtype in (MT.META, MT.SEMANTIC, MT.CONTENT, MT.DATA, MT.INSTANCE, MT.ID, MT.FLAKE):
        encoded_length = encode_length(mtype, bit_length)
    elif mtype == MT.ISCC:
        raise ValueError(f"{mtype} is not a unit")
    else:
        raise ValueError(f"Illegal MainType {mtype}")

    nbytes = bit_length // 8
    header = encode_header(mtype, stype, version, encoded_length)
    body = digest[:nbytes]
    component_code = encode_base32(header + body)
    return component_code


def encode_header(mtype, stype, version=0, length=1):
    # type: (MainType, SubType, Version, Length) -> bytes
    """
    Encodes header values with nibble-sized (4-bit) variable-length encoding.
    The result is minimum 2 and maximum 8 bytes long. If the final count of nibbles
    is uneven it is padded with 4-bit `0000` at the end.

    !!! warning
        The length value must be encoded beforhand because its semantics depend on
        the MainType (see `encode_length` function).

    :param MainType mtype: MainType of unit.
    :param SubType stype: SubType of unit.
    :param Version version: Version of component algorithm.
    :param Length length: length value of unit (1 means 64-bits for standard units)
    :return: Varnibble stream encoded ISCC header as bytes.
    :rtype: bytes

    """
    # TODO verify that all header params and there combination is valid
    header = bitarray()
    for n in (mtype, stype, version, length):
        header += encode_varnibble(n)
    # Append zero-padding if required (right side, least significant bits).
    header.fill()
    return header.tobytes()


def encode_varnibble(n):
    # type: (int) -> bitarray
    """
    Writes integer to variable length sequence of 4-bit chunks.

    Variable-length encoding scheme:

    ------------------------------------------------------
    | prefix bits | nibbles | data bits | unsigned range |
    | ----------- | ------- | --------- | -------------- |
    | 0           | 1       | 3         | 0 - 7          |
    | 10          | 2       | 6         | 8 - 71         |
    | 110         | 3       | 9         | 72 - 583       |
    | 1110        | 4       | 12        | 584 - 4679     |

    :param int n: Positive integer to be encoded as varnibble (0-4679)
    :return: Varnibble encoded integera
    :rtype: bitarray
    """
    if 0 <= n < 8:
        return int2ba(n, length=4)
    elif 8 <= n < 72:
        return bitarray("10") + int2ba(n - 8, length=6)
    elif 72 <= n < 584:
        return bitarray("110") + int2ba(n - 72, length=9)
    elif 584 <= n < 4680:
        return bitarray("1110") + int2ba(n - 584, length=12)
    else:
        raise ValueError("Value must be between 0 and 4679")


def decode_header(data):
    # type: (bytes) -> IsccTuple
    """
    Decodes varnibble encoded header and returns it together with `tail data`.

    Tail data is included to enable decoding of sequential ISCCs. The returned tail
    data must be truncated to decode_length(r[0], r[3]) bits to recover the actual
    hash-bytes.

    :param bytes data: ISCC bytes
    :return: (MainType, SubType, Version, length, TailData)
    :rtype: IsccTuple
    """
    result = []
    ba = bitarray()
    ba.frombytes(data)
    data = ba
    for _ in range(4):
        value, data = decode_varnibble(data)
        result.append(value)

    # Strip 4-bit padding if required
    if len(data) % 8 and data[:4] == bitarray("0000"):
        data = data[4:]

    result.append(data.tobytes())

    return tuple(result)


def decode_varnibble(b):
    # type: (bitarray) -> Tuple[int, bitarray]
    """Reads first varnibble, returns its integer value and remaining bits.

    :param bitarray b: Array of header bits
    :return: A tuple of the integer value of first varnible and the remaining bits.
    :rtype: Tuple[int, bitarray]
    """

    bits = len(b)

    if bits >= 4 and b[0] == 0:
        return ba2int(b[:4]), b[4:]
    if bits >= 8 and b[1] == 0:
        return ba2int(b[2:8]) + 8, b[8:]
    if bits >= 12 and b[2] == 0:
        return ba2int(b[3:12]) + 72, b[12:]
    if bits >= 16 and b[3] == 0:
        return ba2int(b[4:16]) + 584, b[16:]

    raise ValueError("Invalid bitarray")


def encode_units(units):
    # type: (Tuple[MT, ...]) -> int
    """
    Encodes a combination of ISCC units to an integer between 0-7 to be used as length
    value for the final encoding of MT.ISCC

    :param Tuple units: A tuple of a MainType combination (can be empty)
    :return: Integer value to be used as length-value for header encoding
    :rtype: int
    """
    return UNITS.index(units)


def decode_units(unit_id):
    # type: (int) -> Tuple[MT, ...]
    """
    Decodes an ISCC header length value that has been encoded with a unit_id to an
    ordered tuple of MainTypes.
    """
    units = sorted(UNITS[unit_id])
    return tuple(MT(u) for u in units)


def encode_length(mtype, length):
    # type: (MainType, Length) -> int
    """
    Encode length to integer value for header encoding.

    The `length` value has MainType-specific semantics:

    For MainTypes `META`, `SEMANTIC`, `CONTENT`, `DATA`, `INSTANCE`:

        Length means number of bits for the body.
        Length is encoded as the multiple of 32-bit chunks (0 being 32bits)
        Examples: 32 -> 0, 64 -> 1, 96 -> 2 ...

    For MainType `ISCC`:

        MainTypes `DATA` and `INSTANCE` are mandatory for ISCC-CODEs, all others are
        optional. Length means the composition of optional 64-bit units included
        in the ISCC composite.

        Examples:
            No optional units      -> 0000 -> 0
            CONTENT                -> 0001 -> 1
            SEMANTIC               -> 0010 -> 2
            SEMANTIC, CONTENT      -> 0011 -> 3
            META                   -> 0100 -> 4
            META, CONTENT          -> 0101 -> 5
            ...

    For MainType `ID`:

        Lengths means number the number of bits for the body including the counter
        Length is encoded as number of bytes of the counter (64-bit body is implicit)
        Examples:
            64 -> 0 (No counter)
            72 -> 1 (One byte counter)
            80 -> 2 (Two byte counter)
            ...

    :param MainType mtype: The MainType for which to encode the length value.
    :param Length length: The length expressed according to the semantics of the type
    :return: The length value encoded as integer for use with write_header.
    :rtype: int
    """

    error = f"Invalid length {length} for MainType {mtype}"
    # standard case (length field denotes number of 32-bit chunks, 0 being 32-bits)
    if mtype in (MT.META, MT.SEMANTIC, MT.CONTENT, MT.DATA, MT.INSTANCE, MT.FLAKE):
        if length >= 32 and not length % 32:
            return (length // 32) - 1
        raise ValueError(error)
    # flag type encoding of included components (pass through as encoded out-of-band)
    elif mtype == MT.ISCC:
        if 0 <= length <= 7:
            return length
        raise ValueError(error)
    # counter byte lenght encoding
    elif mtype == MT.ID:
        if 64 <= length <= 96:
            return (length - 64) // 8
        raise ValueError(error)
    else:
        raise ValueError(error)


def decode_length(mtype, length):
    # type: (MainType, Length) -> LN
    """
    Dedoce raw length value from ISCC header to length of digest in number of bits.

    Decodes a raw header integer value in to its semantically meaningfull value (eg.
    number of bits)
    """
    if mtype in (MT.META, MT.SEMANTIC, MT.CONTENT, MT.DATA, MT.INSTANCE, MT.FLAKE):
        return LN((length + 1) * 32)
    elif mtype == MT.ISCC:
        return LN(len(decode_units(length)) * 64 + 128)
    elif mtype == MT.ID:
        return LN(length * 8 + 64)
    else:
        raise ValueError(f"Invalid length {length} for MainType {mtype}")


def encode_base32(data):
    # type: (bytes) -> str
    """
    Standard RFC4648 base32 encoding without padding.
    """
    return b32encode(data).decode("ascii").rstrip("=")


def decode_base32(code):
    # type: (str) -> bytes
    """
    Standard RFC4648 base32 decoding without padding and with casefolding.
    """
    # python stdlib does not support base32 without padding, so we have to re-pad.
    cl = len(code)
    pad_length = math.ceil(cl / 8) * 8 - cl

    return bytes(b32decode(code + "=" * pad_length, casefold=True))


def encode_base64(data: bytes) -> str:
    """
    Standard RFC4648 base64url encoding without padding.
    """
    code = urlsafe_b64encode(data).decode("ascii")
    return code.rstrip("=")


def decode_base64(code: str) -> bytes:
    """
    Standard RFC4648 base64url decoding without padding.
    """
    padding = 4 - (len(code) % 4)
    string = code + ("=" * padding)
    return urlsafe_b64decode(string)


def encode_base32hex(data):
    """
    RFC4648 Base32hex encoding without padding

    see: https://tools.ietf.org/html/rfc4648#page-10
    """
    return base32hexnopad.encode(data)


def decode_base32hex(data):
    """
    RFC4648 Base32hex decoding without padding

    see: https://tools.ietf.org/html/rfc4648#page-10
    """
    return base32hexnopad.decode(data)


def iscc_decompose(iscc_code):
    # type: (str) -> List[str]
    """
    Decompose a normalized ISCC-CODE or any valid ISCC sequence into a list of ISCC-UNITS.

    A valid ISCC sequence is a string concatenation of ISCC-UNITS optionally seperated
    by a hyphen.
    """
    iscc_code = iscc_clean(iscc_code)
    components = []

    raw_code = decode_base32(iscc_code)
    while raw_code:
        mt, st, vs, ln, body = decode_header(raw_code)
        # standard ISCC-UNIT with tail continuation
        if mt != MT.ISCC:
            ln_bits = decode_length(mt, ln)
            code = encode_component(mt, st, vs, ln_bits, body[: ln_bits // 8])
            components.append(code)
            raw_code = body[ln_bits // 8 :]
            continue

        # ISCC-CODE
        main_types = decode_units(ln)

        # rebuild dynamic units (META, SEMANTIC, CONTENT)
        for idx, mtype in enumerate(main_types):
            stype = ST.NONE if mtype == MT.META else st
            code = encode_component(mtype, stype, vs, 64, body[idx * 8 :])
            components.append(code)

        # rebuild static units (DATA, INSTANCE)
        data_code = encode_component(MT.DATA, ST.NONE, vs, 64, body[-16:-8])
        instance_code = encode_component(MT.INSTANCE, ST.NONE, vs, 64, body[-8:])
        components.extend([data_code, instance_code])
        break

    return components


def iscc_normalize(iscc_code):
    # type: (str) -> str
    """
    Normalize an ISCC to its canonical form.

    The canonical form of an ISCC is its shortest base32 encoded representation
    prefixed with the string `ISCC:`.

    Possible valid inputs:

        MEACB7X7777574L6
        ISCC:MEACB7X7777574L6
        fcc010001657fe7cafe9791bb
        iscc:maagztfqttvizpjr
        Iscc:Maagztfqttvizpjr


    !!! info
        A concatenated sequence of codes will be composed into a single ISCC of MainType
        `MT.ISCC` if possible.

    !!! example
        ``` py
        >>> import iscc_core
        >>> iscc_core.iscc_normalize("GAAW2PRCRS5LNVZV-IAAUVACQKXE3V44W")
        'ISCC:KUAG2PRCRS5LNVZVJKAFAVOJXLZZM'

        ```

    :param str iscc_code: Any valid ISCC string
    :return: Normalized ISCC
    :rtype: str
    """
    from iscc_core.iscc_code import gen_iscc_code_v0

    decoders = {
        MULTIBASE.base16.value: bytes.fromhex,  # f
        MULTIBASE.base32.value: decode_base32,  # b
        MULTIBASE.base32hex.value: decode_base32hex,  # v
        MULTIBASE.base58btc.value: base58.b58decode,  # z
        MULTIBASE.base64url.value: decode_base64,  # u
    }

    # Transcode to base32 if <multibase><multicodec> encoded
    multibase_prefix = iscc_code[0]
    if multibase_prefix in decoders.keys():
        decoder = decoders[multibase_prefix]
        decoded = decoder(iscc_code[1:])
        if not decoded.startswith(MC_PREFIX):
            raise ValueError(f"Malformed multiformat codec: {decoded[:2]}")
        iscc_code = encode_base32(decoded[2:])
    else:
        prefix = iscc_code.lstrip("ISCC:").lstrip("iscc:")[:2].upper()
        if prefix not in PREFIXES:
            raise ValueError(f"ISCC starts with invalid prefix {prefix}")

    decomposed = iscc_decompose(iscc_code)
    recomposed = gen_iscc_code_v0(decomposed)["iscc"] if len(decomposed) >= 2 else decomposed[0]
    return f"ISCC:{recomposed}" if not recomposed.startswith("ISCC:") else recomposed


########################################################################################
# Convenience functions and classes                                                    #
########################################################################################


def iscc_decode(iscc):
    # type: (str) -> IsccTuple
    """
    Decode ISCC to an IsccTuple

    :param str iscc: ISCC string
    :return: ISCC decoded to an IsccTuple
    :rtype: IsccTuple
    """
    iscc = iscc_clean(iscc_normalize(iscc))
    data = decode_base32(iscc)
    return decode_header(data)


def iscc_explain(iscc):
    # type: (str) -> str
    """
    Convert ISCC to a human-readable representation

    :param str iscc: ISCC string
    :return: Human-readable representation of ISCC
    :rtype: str
    """
    tid = iscc_type_id(iscc)
    fields = iscc_decode(iscc)
    if fields[0] == MT.ID:
        counter_bytes = fields[-1][8:]
        if counter_bytes:
            counter = uvarint.decode(counter_bytes)
            hex_hash = fields[-1][:8].hex()
            return f"{tid}-{hex_hash}-{counter.integer}"
    hex_hash = fields[-1].hex()
    return f"{tid}-{hex_hash}"


def iscc_type_id(iscc):
    # type: (str) -> str
    """
    Extract and convert ISCC HEADER to a readable Type-ID string.

    Type-ids can be used as names in databases to index ISCC-UNITs seperatly.

    :param str iscc: ISCC string
    :return: Unique Type-ID string
    :rtype: str
    """
    fields = iscc_decode(iscc)
    mtype = MT(fields[0])
    stype = SUBTYPE_MAP[fields[0]](fields[1])

    if mtype == MT.ISCC:
        mtypes = decode_units(fields[3])
        length = "".join([t.name[0] for t in mtypes]) + "DI"
    else:
        length = decode_length(fields[0], fields[3])

    version = VS(fields[2])

    return f"{mtype.name}-{stype.name}-{version.name}-{length}"


def iscc_validate(iscc, strict=True):
    # type: (str, bool) -> bool
    """
    Validate that a given string is a *strictly well-formed* ISCC.

    A *strictly well-formed* ISCC is:

    - an ISCC-CODE or ISCC-UNIT
    - encoded with base32 upper without padding
    - has a valid combination of header values
    - is represented in its canonical URI form

    :param str iscc: ISCC string
    :param bool strict: Raise an exeption if validation fails (default True)
    :return: True if sting is valid else false. (raises ValueError in strict mode)
    :rtype: bool
    """

    # Basic regex validation
    match = re.match("^ISCC:[A-Z2-7]{10,60}$", iscc)
    if not match:
        if strict:
            raise ValueError("ISCC string does not match ^ISCC:[A-Z2-7]{10,60}$")
        else:
            return False

    cleaned = iscc_clean(iscc)
    prefix = cleaned[:2]
    if prefix not in PREFIXES:
        if strict:
            raise ValueError(f"Header starts with invalid sequence {prefix}")
        else:
            return False

    m, s, v, l, t = decode_header(decode_base32(cleaned))
    if v != 0:
        if strict:
            raise ValueError(f"Unknown version {v} in version header")
        else:
            return False

    return True


def iscc_clean(iscc):
    # type: (str) -> str
    """
    Cleanup ISCC string.

    Removes leading scheme, dashes, leading/trailing whitespace.

    :param str iscc: Any valid ISCC string
    :return: Cleaned ISCC string.
    :rtype: str
    """
    split = [part.strip() for part in iscc.strip().split(":")]
    if len(split) == 1:
        code = split[0]
        # remove dashes if not multiformat
        if code[0] not in list(MULTIBASE):
            code = code.replace("-", "")
        return code
    elif len(split) == 2:
        scheme, code = split
        if scheme.lower() != "iscc":
            raise ValueError(f"Invalid scheme: {scheme}")
        return code.replace("-", "")
    else:
        raise ValueError(f"Malformed ISCC string: {iscc}")



__version__ = "1.0.0"
from iscc_core.options import core_opts, conformant_options

# Import full api to toplevel
from iscc_core.conformance import *
from iscc_core.constants import *

from iscc_core.simhash import *
from iscc_core.minhash import *
from iscc_core.wtahash import *
from iscc_core.dct import *
from iscc_core.cdc import *

from iscc_core.iscc_code import *
from iscc_core.iscc_id import *
from iscc_core.code_meta import *
from iscc_core.code_content_text import *
from iscc_core.code_content_image import *
from iscc_core.code_content_audio import *
from iscc_core.code_content_video import *
from iscc_core.code_content_mixed import *
from iscc_core.code_data import *
from iscc_core.code_instance import *
from iscc_core.code_flake import *
from iscc_core.codec import *
from iscc_core.utils import *
from iscc_core.models import *
from iscc_core.check import *



# -*- coding: utf-8 -*-
"""*A multi-component identifier for digital media assets.*

An **ISCC-CODE** can be generated from the concatenation of the digests of the following
five **ISCC-UNITs** together with a single common header:

- [Meta-Code](/units/code_meta/) - Encodes metadata similarity
- Semantic-Code - Encodes semantic content similarity (to be developed)
- [Content-Code](/units/content/) - Encodes syntactic/perceptual similarity
- [Data-Code](/units/code_data/) - Encodes raw bitstream similarity
- [Instance-Code](/units/code_instance/) - Data checksum

The following sequences of ISCC-UNITs are possible:

- Data, Instance
- Content, Data, Instance
- Semantic, Data, Instance
- Content, Semantic, Data, Instance
- Meta, Data, Instance
- Meta, Content, Data, Instance
- Meta, Semantic, Data, Instance
- Meta, Semantic, Content, Data, Instance
"""
from typing import Sequence
from operator import itemgetter
import iscc_core as ic


def gen_iscc_code(codes):
    # type: (Sequence[str]) -> dict
    """
    Combine multiple ISCC-UNITs to a composite ISCC-CODE with a common header using
    the latest standard algorithm.

    :param Sequence[str] codes: A valid sequence of singluar ISCC codes.
    :return: An ISCC object with ISCC-CODE
    :rtype: dict
    """
    return gen_iscc_code_v0(codes)


def gen_iscc_code_v0(codes):
    # type: (Sequence[str]) -> dict
    """
    Combine multiple ISCC-UNITS to an ISCC-CODE with a common header using
    algorithm v0.

    :param Sequence[str] codes: A valid sequence of singluar ISCC-UNITS.
    :return: An ISCC object with ISCC-CODE
    :rtype: dict
    """

    codes = [ic.iscc_clean(code) for code in codes]

    # Check basic constraints
    if len(codes) < 2:
        raise ValueError("Minimum two ISCC units required to generate valid ISCC-CODE")
    for code in codes:
        if len(code) < 16:
            raise ValueError(f"Cannot build ISCC-CODE from units shorter than 64-bits: {code}")

    # Decode units and sort by MainType
    decoded = sorted(
        [ic.decode_header(ic.decode_base32(code)) for code in codes], key=itemgetter(0)
    )
    main_types = tuple(d[0] for d in decoded)
    if main_types[-2:] != (ic.MT.DATA, ic.MT.INSTANCE):
        raise ValueError(f"ISCC-CODE requires at least MT.DATA and MT.INSTANCE units.")

    # Determine SubType (generic mediatype)
    sub_types = [t[1] for t in decoded if t[0] in {ic.MT.SEMANTIC, ic.MT.CONTENT}]
    if len(set(sub_types)) > 1:
        raise ValueError(f"Semantic-Code and Content-Code must be of same SubType")
    st = sub_types.pop() if sub_types else ic.ST_ISCC.SUM if len(codes) == 2 else ic.ST_ISCC.NONE

    # Encode unit combination
    encoded_length = ic.encode_units(main_types[:-2])

    # Collect and truncate unit digests to 64-bit
    digest = b"".join([t[-1][:8] for t in decoded])
    header = ic.encode_header(ic.MT.ISCC, st, ic.VS.V0, encoded_length)

    code = ic.encode_base32(header + digest)
    iscc = "ISCC:" + code
    return dict(iscc=iscc)



# -*- coding: utf-8 -*-
"""*A similarity preserving hash for plain-text content (soft hash).*

The ISCC Text-Code is generated from plain-text that has been extracted
from a media assets.

!!! Warning

    Plain-text extraction from documents in various formats (especially PDF) may
    yield very diffent results depending on the extraction tools being used.
    The [iscc-sdk](https://github.com/iscc/iscc-sdk) uses [Apache Tika](https://tika.apache.org)
    to extract text from documents for Text-Code generation.

**Algorithm overview**

- Apply [`text_collapse`][iscc_core.code_content_text.text_collapse] function to text input
- Count characters of collapsed text
- Apply [`soft_hash_text_v0`][iscc_core.code_content_text.soft_hash_text_v0] to collapsed text
"""
import unicodedata
import xxhash
import iscc_core as ic


__all__ = [
    "gen_text_code",
    "gen_text_code_v0",
    "soft_hash_text_v0",
    "text_collapse",
]


def gen_text_code(text, bits=ic.core_opts.text_bits):
    # type: (str, int) -> dict
    """
    Create an ISCC Text-Code with the latest standard algorithm.

    :param str text: Plain text for Text-Code creation.
    :param int bits: Bit-length (multiple of 32) for ISCC Code Hash (default 64).
    :return: ISCC schema instance with Text-Code and an aditional property `characters`
    :rtype: dict
    """
    return gen_text_code_v0(text, bits)


def gen_text_code_v0(text, bits=ic.core_opts.text_bits):
    # type: (str, int) -> dict
    """
    Create an ISCC Text-Code with algorithm v0.

    !!! note
        Any markup (like HTML tags or markdown) should be removed from the plain-text
        before passing it to this function.

    :param str text: Text for Text-Code creation
    :param int bits: Bit-length of ISCC Code Hash (default 64)
    :return: ISCC schema instance with Text-Code and an aditional property `characters`
    :rtype: dict
    """

    text = text_collapse(text)
    characters = len(text)
    digest = soft_hash_text_v0(text)

    text_code = ic.encode_component(
        mtype=ic.MT.CONTENT,
        stype=ic.ST_CC.TEXT,
        version=ic.VS.V0,
        bit_length=bits,
        digest=digest,
    )

    iscc = "ISCC:" + text_code

    return dict(iscc=iscc, characters=characters)


def soft_hash_text_v0(text):
    # type: (str) -> bytes
    """
    Creates a 256-bit similarity preserving hash for text input with algorithm v0.

    - Slide over text with a
      [`text_ngram_size`][iscc_core.options.CoreOptions.text_ngram_size] wide window
      and create [`xxh32`](https://cyan4973.github.io/xxHash/) hashes
    - Create a [`minhash_256`][iscc_core.minhash.alg_minhash_256] from the hashes generated
      in the previous step.

    !!! note

        Before passing text to this function it must be:

        - stripped of markup
        - normalized
        - stripped of whitespace
        - lowercased

    :param str text: Plain text to be hashed.
    :return: 256-bit similarity preserving byte hash.
    :rtype: bytes
    """
    ngrams = ic.sliding_window(text, ic.core_opts.text_ngram_size)
    features = [xxhash.xxh32_intdigest(s.encode("utf-8")) for s in ngrams]
    hash_digest = ic.alg_minhash_256(features)
    return hash_digest


def text_collapse(text):
    # type: (str) -> str
    """
    Normalize and simplify text for similarity hashing.

    - Decompose with NFD normalization.
    - Remove all whitespace characters and convert text to lower case
    - Filter control characters, marks (diacritics), and punctuation
    - Recombine with NFKC normalization.

    !!! note

        See: [Unicode normalization](https://unicode.org/reports/tr15/).

    :param str text: Plain text to be collapsed.
    :return: Collapsed plain text.
    :rtype: str
    """

    # Decompose with NFD
    text = unicodedata.normalize("NFD", text)

    # Remove all whitespace and convert text to lower case
    text = "".join(text.split()).lower()

    # Filter control characters, marks (diacritics), and punctuation
    text = "".join(
        ch for ch in text if unicodedata.category(ch)[0] not in ic.core_opts.text_unicode_filter
    )

    # Recombine
    text = unicodedata.normalize("NFKC", text)

    return text



# -*- coding: utf-8 -*-
"""*A similarity hash for mixed media content.*

Creates an ISCC object that provides a `iscc`-field a Mixed-Code and a `parts`-field
that lists the input codes.

Many digital assets embed multiple assets of different mediatypes in a single file.
Text documents may include images, video includes audio in most cases. The ISCC
Content-Code-Mixed encodes the similarity of a collection of assets of the same or
different mediatypes that may occur in a multimedia asset.

Applications that create mixed Content-Codes must be capable to extract embedded
assets and create individual Content-Codes per asset.
"""
from typing import Iterable, Sequence
import iscc_core as ic


__all__ = [
    "gen_mixed_code",
    "gen_mixed_code_v0",
    "soft_hash_codes_v0",
]


def gen_mixed_code(codes, bits=ic.core_opts.mixed_bits):
    # type: (Sequence[str], int) -> dict
    """
    Create an ISCC Content-Code Mixed with the latest standard algorithm.

    :param Iterable[str] codes: a list of Content-Codes.
    :param int bits: Target bit-length of generated Content-Code-Mixed.
    :return: ISCC object with Content-Code Mixed.
    :rtype: dict
    """
    return gen_mixed_code_v0(codes, bits=bits)


def gen_mixed_code_v0(codes, bits=ic.core_opts.mixed_bits):
    # type: (Sequence[str], int) -> dict
    """
    Create an ISCC Content-Code-Mixed with algorithm v0.

    If the provided codes are of mixed length they are stripped to `bits` length for
    calculation.

    :param Iterable[str] codes: a list of Content-Codes.
    :param int bits: Target bit-length of generated Content-Code-Mixed.
    :return: ISCC object with Content-Code Mixed.
    :rtype: dict
    """
    digests = [ic.decode_base32(ic.iscc_clean(code)) for code in codes]
    digest = soft_hash_codes_v0(digests, bits=bits)
    mixed_code = ic.encode_component(
        mtype=ic.MT.CONTENT,
        stype=ic.ST_CC.MIXED,
        version=ic.VS.V0,
        bit_length=bits,
        digest=digest,
    )
    iscc = "ISCC:" + mixed_code
    return dict(iscc=iscc, parts=list(codes))


def soft_hash_codes_v0(cc_digests, bits=ic.core_opts.mixed_bits):
    # type: (Sequence[bytes], int) -> bytes
    """
    Create a similarity hash from multiple Content-Code digests.

    The similarity hash is created from the bodies of the input codes with the first
    byte of the code-header prepended.

    All codes must be of main-type CONTENT and have a minimum length of `bits`.

    :param Sequence[bytes] cc_digests: a list of Content-Code digests.
    :param int bits: Target bit-length of generated Content-Code-Mixed.
    :return: Similarity preserving byte hash.
    :rtype: bytes
    """
    if not len(cc_digests) > 1:
        raise AssertionError("Minimum of 2 codes needed for Content-Code-Mixed.")
    nbytes = bits // 8
    code_tuples = [ic.decode_header(code) for code in cc_digests]
    if not all([ct[0] == ic.MT.CONTENT for ct in code_tuples]):
        raise AssertionError(
            "Only codes with main-type CONTENT allowed as input for Content-Code-Mixed"
        )

    unit_lengths = [ic.decode_length(t[0], t[3]) for t in code_tuples]
    if not all(ul >= bits for ul in unit_lengths):
        raise AssertionError(f"Code to short for {bits}-bit length")

    hash_bytes = []
    # Retain the first byte of the header and strip body to mixed_bits length
    for full, code_tuple in zip(cc_digests, code_tuples):
        hash_bytes.append(full[:1] + code_tuple[-1][: nbytes - 1])
    return ic.alg_simhash(hash_bytes)



# -*- coding: utf-8 -*-
import enum
from typing import Tuple, Union


########################################################################################
# Type definitions and constants                                                       #
########################################################################################


Data = Union[bytes, bytearray, memoryview]
Stream = Union["BinaryIO", "mmap.mmap", "BytesIO", "BufferedReader"]
MainType = Union[int, "MT"]
SubType = Union[int, "ST", "ST_CC", "ST_ISCC", "ST_ID"]
Version = Union[int, "VS"]
Length = Union[int, "LN"]
Header = Tuple[MainType, SubType, Version, Length]
IsccTuple = Tuple[MainType, SubType, Version, Length, bytes]
IsccAny = Union[str, IsccTuple, bytes, "Code"]
Meta = Union[dict, str]
FrameSig = Tuple[int]


class MT(enum.IntEnum):
    """
    ## MT - MainTypes

    | Uint | Symbol   | Bits | Purpose                                                 |
    |----- |:---------|------|---------------------------------------------------------|
    | 0    | META     | 0000 | Match on metadata similarity                            |
    | 1    | SEMANTIC | 0001 | Match on semantic content similarity                    |
    | 2    | CONTENT  | 0010 | Match on perceptual content similarity                  |
    | 3    | DATA     | 0011 | Match on data similarity                                |
    | 4    | INSTANCE | 0100 | Match on data identity                                  |
    | 5    | ISCC     | 0101 | Composite of two or more ISCC-UNITs with common header  |
    """

    META = 0
    SEMANTIC = 1
    CONTENT = 2
    DATA = 3
    INSTANCE = 4
    ISCC = 5
    ID = 6
    FLAKE = 7


class ST(enum.IntEnum):
    """
    ## ST - SubTypes

    | Uint | Symbol   | Bits | Purpose                                                 |
    |----- |:---------|------|---------------------------------------------------------|
    | 0    | NONE     | 0000 | For MainTypes that do not specify SubTypes              |
    """

    NONE = 0


class ST_CC(enum.IntEnum):
    """
    ### ST_CC

    SubTypes for `MT.CONTENT`

    | Uint | Symbol   | Bits | Purpose                                                 |
    |----- |:---------|------|---------------------------------------------------------|
    | 0    | TEXT     | 0000 | Match on syntactic text similarity                      |
    | 1    | IMAGE    | 0001 | Match on perceptual image similarity                    |
    | 2    | AUDIO    | 0010 | Match on audio chroma similarity                        |
    | 3    | VIDEO    | 0011 | Match on perceptual similarity                          |
    | 4    | MIXED    | 0100 | Match on similarity of content codes                    |
    """

    TEXT = 0
    IMAGE = 1
    AUDIO = 2
    VIDEO = 3
    MIXED = 4


class ST_ISCC(enum.IntEnum):
    """
    ### ST_ISCC

    SubTypes for `MT.ISCC`

    | Uint | Symbol   | Bits | Purpose                                                 |
    |----- |:---------|------|---------------------------------------------------------|
    | 0    | TEXT     | 0000 | Composite ISCC inlcuding Text-Code                      |
    | 1    | IMAGE    | 0001 | Composite ISCC inlcuding Image-Code                     |
    | 2    | AUDIO    | 0010 | Composite ISCC inlcuding Audio-Code                     |
    | 3    | VIDEO    | 0011 | Composite ISCC inlcuding Video-Code                     |
    | 4    | MIXED    | 0100 | Composite ISCC inlcuding Mixed-Code                     |
    | 5    | SUM      | 0101 | Composite ISCC inlcuding only Data- and Instance-Code   |
    | 6    | NONE     | 0110 | Composite ISCC including Meta, Data and Instance-Code   |
    """

    TEXT = 0
    IMAGE = 1
    AUDIO = 2
    VIDEO = 3
    MIXED = 4
    SUM = 5
    NONE = 6


class ST_ID(enum.IntEnum):
    """
    ### ST_ID

    SubTypes for `MT.ID`

    | Uint | Symbol   | Bits | Purpose                                                 |
    |----- |:---------|------|---------------------------------------------------------|
    | 0    | PRIVATE  | 0000 | ISCC-ID minted via private repository (not unique)      |
    | 1    | BITCOIN  | 0001 | ISCC-ID minted via Bitcoin blockchain                   |
    | 2    | ETHEREUM | 0010 | ISCC-ID minted via Ethereum blockchain                  |
    | 3    | POLYGON  | 0011 | ISCC-ID minted via Polygon blockchain                   |
    """

    PRIVATE = 0
    BITCOIN = 1
    ETHEREUM = 2
    POLYGON = 3


class VS(enum.IntEnum):
    """
    ## VS - Version

    Code Version

    | Uint | Symbol   | Bits | Purpose                                                 |
    |----- |:---------|------|---------------------------------------------------------|
    | 0    | V0       | 0000 | Initial Version of Code without breaking changes        |

    """

    V0 = 0


class LN(enum.IntEnum):
    """
    ## LN - Length

    Valid lengths for hash-digests.

    - L32 = 32
    - L64 = 64
    - L72 = 72
    - L80 = 80
    - L96 = 96
    - L128 = 128
    - L160 = 160
    - L192 = 192
    - L224 = 224
    - L256 = 256
    """

    L32 = 32
    L64 = 64
    L72 = 72
    L80 = 80
    L96 = 96
    L128 = 128
    L160 = 160
    L192 = 192
    L224 = 224
    L256 = 256


class MULTIBASE(str, enum.Enum):
    """
    Supported Multibase encodings.

    - base16 -> f
    - base32 -> b
    - base32hex -> v
    - base58btc -> z
    - base64url -> u
    """

    base16 = "f"
    base32 = "b"
    base32hex = "v"
    base58btc = "z"
    base64url = "u"


#: Possible combinations of ISCC units for the first 3 units of MT.ISCC
UNITS = (
    tuple(),
    (MT.CONTENT,),
    (MT.SEMANTIC,),
    (MT.CONTENT, MT.SEMANTIC),
    (MT.META,),
    (MT.META, MT.CONTENT),
    (MT.META, MT.SEMANTIC),
    (MT.META, MT.SEMANTIC, MT.CONTENT),
)

#: Map MainTypes to SubTypes
SUBTYPE_MAP = {
    MT.META: ST,
    MT.SEMANTIC: ST_CC,
    MT.CONTENT: ST_CC,
    MT.DATA: ST,
    MT.INSTANCE: ST,
    MT.ISCC: ST_ISCC,
    MT.ID: ST_ID,
    MT.FLAKE: ST,
}

#: Multicodec prefix code
MC_PREFIX: bytes = b"\xcc\x01"

PREFIXES = [
    "AA",
    "CA",
    "CE",
    "CI",
    "CM",
    "CQ",
    "EA",
    "EE",
    "EI",
    "EM",
    "EQ",
    "GA",
    "IA",
    "KA",
    "KE",
    "KI",
    "KM",
    "KQ",
    "KU",
    "KY",
    "MA",
    "ME",
    "MI",
    "MM",
    "OA",
]



# -*- coding: utf-8 -*-
"""Inspect lib environment/installation"""
import importlib.machinery
import inspect


__all__ = ["turbo"]

EXTENSION_SUFFIXES = tuple(sf.lstrip(".") for sf in importlib.machinery.EXTENSION_SUFFIXES)


def suffix(filename):
    return "." in filename and filename.rpartition(".")[-1] or ""


def isnativemodule(module):
    """isnativemodule(thing) → boolean predicate, True if `module`
    is a native-compiled (“extension”) module.

    Q.v. this fine StackOverflow answer on this subject:
        https://stackoverflow.com/a/39304199/298171
    """
    # Step one: modules only beyond this point:
    if not inspect.ismodule(module):  # pragma: no cover
        return False

    # Step two: return truly when “__loader__” is set:
    if isinstance(
        getattr(module, "__loader__", None), importlib.machinery.ExtensionFileLoader
    ):  # pragma: no cover
        return True

    # Step three: in leu of either of those indicators,
    # check the module path’s file suffix:
    try:
        ext = suffix(inspect.getfile(module))
    except TypeError as exc:
        return "is a built-in" in str(exc)

    return ext in EXTENSION_SUFFIXES


def turbo():
    # type: () -> bool
    """Check whether all optional cython extensions have been compiled to native modules."""
    from iscc_core import cdc, minhash, simhash

    modules = (cdc, minhash, simhash)
    for module in modules:
        if not isnativemodule(module):
            return False
    return True  # pragma: no cover



# -*- coding: utf-8 -*-
import math
from typing import List, Sequence


def alg_dct(v):
    # type: (Sequence[float]) -> List
    """
    Discrete cosine transform.

    See: [nayuki.io](https://www.nayuki.io/page/fast-discrete-cosine-transform-algorithms).

    :param Sequence[float] v: Input vector for DCT calculation.
    :return: DCT Transformed vector.
    :rtype: List
    """

    n = len(v)
    if n == 1:
        return list(v)
    elif n == 0 or n % 2 != 0:
        raise ValueError()
    else:
        half = n // 2
        alpha = [(v[i] + v[-(i + 1)]) for i in range(half)]
        beta = [
            (v[i] - v[-(i + 1)]) / (math.cos((i + 0.5) * math.pi / n) * 2.0) for i in range(half)
        ]
        alpha = alg_dct(alpha)
        beta = alg_dct(beta)
        result = []
        for i in range(half - 1):
            result.append(alpha[i])
            result.append(beta[i] + beta[i + 1])
        result.append(alpha[-1])
        result.append(beta[-1])
        return result



# -*- coding: utf-8 -*-
import io
import json
from hashlib import sha256
from typing import Generator, Sequence, Tuple, Any
import uvarint
from bitarray import bitarray
from bitarray.util import count_xor
from blake3 import blake3
import iscc_core as ic
import jcs
from iscc_core.constants import Stream


__all__ = [
    "json_canonical",
    "cidv1_hex",
    "cidv1_to_token_id",
    "cidv1_from_token_id",
    "sliding_window",
    "iscc_similarity",
    "iscc_compare",
    "iscc_distance",
    "iscc_distance_bytes",
    "multi_hash_blake3",
]


def json_canonical(obj):
    # type: (Any) -> bytes
    """
    Canonical, deterministic serialization of ISCC metadata.

    We serialize ISCC metadata in a deterministic/reproducible manner by using
    [JCS (RFC 8785)](https://datatracker.ietf.org/doc/html/rfc8785) canonicalization.
    """
    ser = jcs.canonicalize(obj)
    des = json.loads(ser)
    if des != obj:
        raise ValueError(f"Not canonicalizable {obj} round-trips to {des}")
    return ser


def cidv1_hex(stream):
    # type: (Union[Stream, bytes]) -> str
    """
    Create a [IPFS CIDv1](https://specs.ipld.io/block-layer/CID.html#cids-version-1) hash for ISCC
    metadata in `base16` (hexadecimal) representation.

    Learn more about IPFS CIDv1 at [ProtoSchool](https://proto.school/anatomy-of-a-cid)

    We use the default `CIDv1` with `sha1-256` and chunksize `262144` as hashing
    algorithm for ISCC metadata.

    !!! attention
        We use the `base16` (hexadecimal) representation for the CIDv1. This gives as
        stable, fixed-length CIDv1 header-prefix and allows us to use the CIDv1 as an
        [ERC-721](https://eips.ethereum.org/EIPS/eip-721)/[ERC-1155](https://eips.ethereum.org/EIPS/eip-1155)
        `uint256` Token-ID while also supporting ID-substitution for the metadata URI.

        For details see discussion at
        [OpenZeppelin Forum](https://forum.openzeppelin.com/t/how-to-erc-1155-id-substitution-for-token-uri/3312/14)


    !!! example
        ```python
        >>> import io
        >>> import iscc_core as ic
        >>> ic.cidv1_hex(io.BytesIO(b'hello world'))
        'f01551220b94d27b9934d3e08a52e52d7da7dabfac484efe37a5380ee9088f7ace2efcde9'

        ```

        The result might not look like a valid IPFS CIDv1, but it actually is:
        https://ipfs.io/ipfs/f01551220b94d27b9934d3e08a52e52d7da7dabfac484efe37a5380ee9088f7ace2efcde9

        With IPFS v0.11.0 this equals to:
        ```bash
        $ipfs add --cid-version=1 <myfile>
        $ipfs cid format -b=base16 bafkreifzjut3te2nhyekklss27nh3k72ysco7y32koao5eei66wof36n5e
        f01551220b94d27b9934d3e08a52e52d7da7dabfac484efe37a5380ee9088f7ace2efcde9
        ```

        When using `cidv1_to_token_id` for this hash you get an `uint256` that can be converted
        to a hex representation within a Smart Contract. The hex-representation can then be
        used for ID-Substitution. Use `ipfs://f01551220{id}` as template for the Metadata-URI.

    :param Stream stream: Data to be hashed (currently max 262144 bytes supported)
    :return: A valid IPFS CIDv1 that can be used as token-id and metadata-uri
    :rtype: str
    """

    if isinstance(stream, bytes):
        stream = io.BytesIO(stream)

    ipfs_max_size = 262144
    data = stream.read(ipfs_max_size)

    # fail if we have more data than ipfs_max_size
    if stream.read(1):
        raise ValueError(
            f"Data exceeds current max size {ipfs_max_size} for ipfs_hash: {len(data)}"
        )

    digest = sha256(data).digest()
    multibase_prefix = "f"
    multicodec_cidv1 = b"\x01"
    multicodec_content_type = b"\x55"  # raw
    multicodec_mutihash_type = b"\x12"  # sha2-256
    multicodec_mutihash_len = uvarint.encode(32)  # 28 byte length (varint encoded 0x1c)

    cid_v1_digest = b"".join(
        (
            multicodec_cidv1,
            multicodec_content_type,
            multicodec_mutihash_type,
            multicodec_mutihash_len,
            digest,
        )
    )

    cid_v1 = multibase_prefix + cid_v1_digest.hex()
    return cid_v1


def cidv1_to_token_id(cidv1):
    # type: (str) -> int
    """
    Convert default IPFS CIDv1 to an uint256 Token-ID.

    To save storage space in Smart Contracts and to securely link an NFT Token-ID to its associated
    metadata we can convert a CIDv1 to an uint256 Token-ID and vice versa.

    !!! example

        ```python
        >>> import io
        >>> import iscc_core as ic
        >>> ic.cidv1_to_token_id("f01551220b94d27b9934d3e08a52e52d7da7dabfac484efe37a5380ee9088f7ace2efcde9")
        83814198383102558219731078260892729932246618004265700685467928187377105751529

        ```

    :param str cidv1: An IPFS CIDv1 string base-16 (hex) encoded representation.
    :return int: An uint256 derived from the CIDv1
    """
    if not cidv1.startswith("f"):
        raise ValueError(f"Only base16 encoded CIDv1 supported. Got {cidv1}")
    nobase = cidv1[1:]
    decoded = bytes.fromhex(nobase)
    if not decoded.startswith(b"\x01\x55\x12\x20"):
        raise ValueError(f"Only sha2-256 with raw leaves supported. Got {decoded.hex()}")
    digest = decoded[4:]
    if not len(digest) == 32:
        raise ValueError(f"Illegal digest size {len(digest)} for sha2-256")
    return int.from_bytes(digest, "big", signed=False)


def cidv1_from_token_id(token_id):
    # type: (int) -> str
    """
    Convert Token-ID to default IPFS CIDv1 (reverse of `cidv1_to_token_id`).

    !!! example
        ```python
        >>> import io
        >>> import iscc_core as ic
        >>> ic.cidv1_from_token_id(83814198383102558219731078260892729932246618004265700685467928187377105751529)
        'f01551220b94d27b9934d3e08a52e52d7da7dabfac484efe37a5380ee9088f7ace2efcde9'

        ```

    :param token_id: An uint256 Token-ID derived from a CIDv1
    :return str: An IPFS CIDv1 string with default base32 encoding.
    """
    digest = b"\x01\x55\x12\x20" + int.to_bytes(token_id, length=32, byteorder="big", signed=False)
    return "f" + digest.hex()


def multi_hash_blake3(data):
    # type: (bytes) -> str
    """
    Create blake3 hash with multihash prefix.

    :param bytes data: Bytes to be hashed
    :return: Multihash prefixed 256-bit blake3 hash as hex string
    :rtype: str
    """
    return (b"\x1e\x20" + blake3(data).digest()).hex()


def sliding_window(seq, width):
    # type: (Sequence, int) -> Generator
    """
    Generate a sequence of equal "width" slices each advancing by one elemnt.

    All types that have a length and can be sliced are supported (list, tuple, str ...).
    The result type matches the type of the input sequence.
    Fragment slices smaller than the width at the end of the sequence are not produced.
    If "witdh" is smaller than the input sequence than one element will be returned that
    is shorter than the requested width.

    :param Sequence seq: Sequence of values to slide over
    :param int width: Width of sliding window in number of items
    :returns: A generator of window sized items
    :rtype: Generator
    """
    if width < 2:
        raise AssertionError("Sliding window width must be 2 or bigger.")
    idx = range(max(len(seq) - width + 1, 1))
    return (seq[i : i + width] for i in idx)


def iscc_compare(a, b):
    # type: (str, str) -> dict
    """
    Calculate separate hamming distances of compatible components of two ISCCs

    :return: A dict with keys meta_dist, semantic_dist, content_dist, data_dist, instance_match
    :rtype: dict
    """
    ac = [ic.Code(unit) for unit in ic.iscc_decompose(a)]
    bc = [ic.Code(unit) for unit in ic.iscc_decompose(b)]
    result = {}
    for ca in ac:
        for cb in bc:
            cat = (ca.maintype, ca.subtype, ca.version)
            cbt = (cb.maintype, cb.subtype, ca.version)
            if cat == cbt:
                if ca.maintype != ic.MT.INSTANCE:
                    result[ca.maintype.name.lower() + "_dist"] = iscc_distance_bytes(
                        ca.hash_bytes, cb.hash_bytes
                    )
                else:
                    result["instance_match"] = ca.hash_bytes == cb.hash_bytes
    return result


def iscc_similarity(a, b):
    # type: (str, str) -> int
    """
    Calculate similarity of ISCC codes as a percentage value (0-100).

    MainType, SubType, Version and Length of the codes must be the same.

    :param a: ISCC a
    :param b: ISCC b
    :return: Similarity of ISCC a and b in percent (based on hamming distance)
    :rtype: int
    """
    a, b = iscc_pair_unpack(a, b)
    hdist = iscc_distance_bytes(a, b)
    nbits = len(a) * 8
    sim = int(((nbits - hdist) / nbits) * 100)
    return sim


def iscc_distance(a, b):
    # type: (str, str) -> int
    """
    Calculate hamming distance of ISCC codes.

    MainType, SubType, Version and Length of the codes must be the same.

    :param a: ISCC a
    :param b: ISCC b
    :return: Hamming distanced in number of bits.
    :rtype: int
    """
    a, b = iscc_pair_unpack(a, b)
    return iscc_distance_bytes(a, b)


def iscc_distance_bytes(a, b):
    # type: (bytes, bytes) -> int
    """
    Calculate hamming distance for binary hash digests of equal length.

    :param bytes a: binary hash digest
    :param bytes b: binary hash digest
    :return: Hamming distance in number of bits.
    :rtype: int
    """
    if len(a) != len(b):
        raise AssertionError(f"Hash diggest of unequal length: {len(a)} vs {len(b)}")
    ba, bb = bitarray(), bitarray()
    ba.frombytes(a)
    bb.frombytes(b)
    return count_xor(ba, bb)


def iscc_pair_unpack(a, b):
    # type: (str, str) -> Tuple[bytes, bytes]
    """
    Unpack two ISCC codes and return their body hash digests if their headers match.

    Headers match if their MainType, SubType, and Version are identical.

    :param a: ISCC a
    :param b: ISCC b
    :return: Tuple with hash digests of a and b
    :rtype: Tuple[bytes, bytes]
    :raise ValueError: If ISCC headers don´t match
    """
    a, b = ic.iscc_clean(ic.iscc_normalize(a)), ic.iscc_clean(ic.iscc_normalize(b))
    a, b = ic.decode_base32(a), ic.decode_base32(b)
    a, b = ic.decode_header(a), ic.decode_header(b)
    if not a[:-1] == b[:-1]:
        raise ValueError(f"ISCC headers don´t match: {a}, {b}")
    return a[-1], b[-1]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iscc_core']

package_data = \
{'': ['*']}

install_requires = \
['base58>=2.1,<3.0',
 'bases>=0.2,<0.3',
 'bitarray-hardbyte>=2.3,<3.0',
 'blake3>=0.3,<0.4',
 'data-url>=1.0,<2.0',
 'jcs>=0.2,<0.3',
 'loguru>=0.6,<0.7',
 'more-itertools>=9.0,<10.0',
 'pybase64>=1.2,<2.0',
 'pydantic[dotenv]',
 'uvarint>=1.2,<2.0',
 'xxhash>=3.0,<4.0']

extras_require = \
{'turbo': ['cython>=0.29,<0.30']}

setup_kwargs = {
    'name': 'iscc-core',
    'version': '1.0.0',
    'description': 'ISCC - Core Algorithms',
    'long_description': '# ISCC - Codec & Algorithms\n\n[![Build](https://github.com/iscc/iscc-core/actions/workflows/tests.yml/badge.svg)](https://github.com/iscc/iscc-core/actions/workflows/tests.yml)\n[![Version](https://img.shields.io/pypi/v/iscc-core.svg)](https://pypi.python.org/pypi/iscc-core/)\n[![Coverage](https://codecov.io/gh/iscc/iscc-core/branch/main/graph/badge.svg?token=7BJ7HJU815)](https://codecov.io/gh/iscc/iscc-core)\n[![Quality](https://app.codacy.com/project/badge/Grade/ad1cc48ac0c0413ea2373a938148f019)](https://www.codacy.com/gh/iscc/iscc-core/dashboard)\n[![Downloads](https://pepy.tech/badge/iscc-core)](https://pepy.tech/project/iscc-core)\n\n`iscc-core` is the reference implementation of the core algorithms of the [ISCC](https://iscc.codes)\n(*International Standard Content Code*)\n\n## What is the ISCC\n\nThe ISCC is a similarity preserving fingerprint and identifier for digital media assets.\n\nISCCs are generated algorithmically from digital content, just like cryptographic hashes. However,\ninstead of using a single cryptographic hash function to identify data only, the ISCC uses various\nalgorithms to create a composite identifier that exhibits similarity-preserving properties (soft\nhash).\n\nThe component-based structure of the ISCC identifies content at multiple levels of abstraction. Each\ncomponent is self-describing, modular, and can be used separately or with others to aid in various\ncontent identification tasks. The algorithmic design supports content deduplication, database\nsynchronization, indexing, integrity verification, timestamping, versioning, data provenance,\nsimilarity clustering, anomaly detection, usage tracking, allocation of royalties, fact-checking and\ngeneral digital asset management use-cases.\n\n## What is `iscc-core`\n\n`iscc-core` is a python based reference library of the core algorithms to create standard-compliant\nISCC codes. It also a good reference for porting ISCC to other programming languages.\n\n!!! tip\n    This is a low level reference implementation that does not inlcude features like mediatype\n    detection, metadata extraction or file format specific content extraction. Please have a look at\n    the [iscc-sdk](https://github.com/iscc/iscc-sdk) which adds those higher level features on top\n    of the `iscc-core` library.\n\n## Project Status\n\nThe ISCC is under development as [ISO/CD 24138](https://www.iso.org/standard/77899.html) -\nInternational Standard Content Code within\n[ISO/TC 46/SC 9/WG 18](https://www.iso.org/committee/48836.html).\n\n## ISCC Architecture\n\n![ISCC Architecture](https://raw.githubusercontent.com/iscc/iscc-core/master/docs/images/iscc-codec-light.png)\n\n## ISCC MainTypes\n\n| Idx | Slug     | Bits | Purpose                                                |\n| --- | :------- | ---- | ------------------------------------------------------ |\n| 0   | META     | 0000 | Match on metadata similarity                           |\n| 1   | SEMANTIC | 0001 | Match on semantic content similarity                   |\n| 2   | CONTENT  | 0010 | Match on perceptual content similarity                 |\n| 3   | DATA     | 0011 | Match on data similarity                               |\n| 4   | INSTANCE | 0100 | Match on data identity                                 |\n| 5   | ISCC     | 0101 | Composite of two or more components with common header |\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install `iscc-core`.\n\n```bash\npip install iscc-core\n```\n\n## Quick Start\n\n```python\nimport json\nimport iscc_core as ic\n\nmeta_code = ic.gen_meta_code(name="ISCC Test Document!")\n\nprint(f"Meta-Code:     {meta_code[\'iscc\']}")\nprint(f"Structure:     {ic.iscc_explain(meta_code[\'iscc\'])}\\n")\n\n# Extract text from file\nwith open("demo.txt", "rt", encoding="utf-8") as stream:\n    text = stream.read()\n    text_code = ic.gen_text_code_v0(text)\n    print(f"Text-Code:     {text_code[\'iscc\']}")\n    print(f"Structure:     {ic.iscc_explain(text_code[\'iscc\'])}\\n")\n\n# Process raw bytes of textfile\nwith open("demo.txt", "rb") as stream:\n    data_code = ic.gen_data_code(stream)\n    print(f"Data-Code:     {data_code[\'iscc\']}")\n    print(f"Structure:     {ic.iscc_explain(data_code[\'iscc\'])}\\n")\n\n    stream.seek(0)\n    instance_code = ic.gen_instance_code(stream)\n    print(f"Instance-Code: {instance_code[\'iscc\']}")\n    print(f"Structure:     {ic.iscc_explain(instance_code[\'iscc\'])}\\n")\n\n# Combine ISCC-UNITs into ISCC-CODE\niscc_code = ic.gen_iscc_code(\n    (meta_code["iscc"], text_code["iscc"], data_code["iscc"], instance_code["iscc"])\n)\n\n# Create convenience `Code` object from ISCC string\niscc_obj = ic.Code(iscc_code["iscc"])\nprint(f"ISCC-CODE:     {ic.iscc_normalize(iscc_obj.code)}")\nprint(f"Structure:     {iscc_obj.explain}")\nprint(f"Multiformat:   {iscc_obj.mf_base32}\\n")\n\n# Compare with changed ISCC-CODE:\nnew_dc, new_ic = ic.Code.rnd(mt=ic.MT.DATA), ic.Code.rnd(mt=ic.MT.INSTANCE)\nnew_iscc = ic.gen_iscc_code((meta_code["iscc"], text_code["iscc"], new_dc.uri, new_ic.uri))\nprint(f"Compare ISCC-CODES:\\n{iscc_obj.uri}\\n{new_iscc[\'iscc\']}")\nprint(json.dumps(ic.iscc_compare(iscc_obj.code, new_iscc["iscc"]), indent=2))\n```\n\nThe output of this example is as follows:\n\n```\nMeta-Code:     ISCC:AAAT4EBWK27737D2\nStructure:     META-NONE-V0-64-3e103656bffdfc7a\n\nText-Code:     ISCC:EAAQMBEYQF6457DP\nStructure:     CONTENT-TEXT-V0-64-060498817dcefc6f\n\nData-Code:     ISCC:GAA7UJMLDXHPPENG\nStructure:     DATA-NONE-V0-64-fa258b1dcef791a6\n\nInstance-Code: ISCC:IAA3Y7HR2FEZCU4N\nStructure:     INSTANCE-NONE-V0-64-bc7cf1d14991538d\n\nISCC-CODE:     ISCC:KACT4EBWK27737D2AYCJRAL5Z36G76RFRMO4554RU26HZ4ORJGIVHDI\nStructure:     ISCC-TEXT-V0-MCDI-3e103656bffdfc7a060498817dcefc6ffa258b1dcef791a6bc7cf1d14991538d\nMultiformat:   bzqavabj6ca3fnp757r5ambeyqf6457dp7isywhoo66i2npd46hiutektru\n\nCompare ISCC-CODES:\nISCC:KACT4EBWK27737D2AYCJRAL5Z36G76RFRMO4554RU26HZ4ORJGIVHDI\nISCC:KACT4EBWK27737D2AYCJRAL5Z36G7Y7HA2BMECKMVRBEQXR2BJOS6NA\n{\n  "meta_dist": 0,\n  "content_dist": 0,\n  "data_dist": 33,\n  "instance_match": false\n}\n```\n\n## Documentation\n\nDocumentation is published at<https://core.iscc.codes>\n\n## Development\n\n**Requirements**\n\n- [Python 3.7.2](https://www.python.org/) or higher for code generation and static site building.\n- [Poetry](https://python-poetry.org/) for installation and dependency management.\n\n**Development Setup**\n\n```shell\ngit clone https://github.com/iscc/iscc-core.git\ncd iscc-core\npoetry install\n```\n\n**Development Tasks**\n\nTests, coverage, code formatting and other tasks can be run with the `poe` command:\n\n```shell\npoe\n\nPoe the Poet - A task runner that works well with poetry.\nversion 0.18.1\n\nResult: No task specified.\n\nUSAGE\n  poe [-h] [-v | -q] [--root PATH] [--ansi | --no-ansi] task [task arguments]\n\nGLOBAL OPTIONS\n  -h, --help     Show this help page and exit\n  --version      Print the version and exit\n  -v, --verbose  Increase command output (repeatable)\n  -q, --quiet    Decrease command output (repeatable)\n  -d, --dry-run  Print the task contents but don\'t actually run it\n  --root PATH    Specify where to find the pyproject.toml\n  --ansi         Force enable ANSI output\n  --no-ansi      Force disable ANSI output\nCONFIGURED TASKS\n  gentests       Generate conformance test data\n  format         Code style formating with black\n  docs           Copy README.md to /docs\n  format-md      Markdown formating with mdformat\n  lf             Convert line endings to lf\n  test           Run tests with coverage\n  sec            Security check with bandit\n  all\n```\n\nUse `poe all` to run all tasks before committing any changes.\n\n## Maintainers\n\n[@titusz](https://github.com/titusz)\n\n## Contributing\n\nPull requests are welcome. For significant changes, please open an issue first to discuss your\nplans. Please make sure to update tests as appropriate.\n\nYou may also want join our developer chat on Telegram at <https://t.me/iscc_dev>.\n',
    'author': 'Titusz',
    'author_email': 'tp@py7.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://iscc.codes',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.2,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""*A decentralized, owned, and short identifier for digital assets.*

The **ISCC-ID** is generated from a similarity-hash of the units of an
**ISCC-CODE** together with a blockchain wallet address. Its SubType designates the blockchain
from which the **ISCC-ID** was minted. The similarity-hash is always at least 64-bits and
optionally suffixed with a [`uvarint`](https://github.com/multiformats/unsigned-varint) endcoded
`uniqueness counter`. The `uniqueness counter` is added and incremented only if the
mint colides with a pre-existing **ISCC-ID** minted from the same blockchain from a
different **ISCC-CODE** or from an identical **ISCC-CODE** registered by a different
signatory.
"""
from hashlib import sha256
from typing import Optional
import uvarint
import iscc_core as ic


__all__ = [
    "gen_iscc_id",
    "gen_iscc_id_v0",
    "iscc_id_incr",
    "iscc_id_incr_v0",
    "alg_simhash_from_iscc_id",
]


def gen_iscc_id(iscc_code, chain_id, wallet, uc=0):
    # type: (str, int, str, Optional[int]) -> dict
    """
    Generate  ISCC-ID from ISCC-CODE with the latest standard algorithm.

    :param str iscc_code: The ISCC-CODE from which to mint the ISCC-ID.
    :param int chain_id: Chain-ID of blockchain from which the ISCC-ID is minted.
    :param str wallet: The wallet address that signes the ISCC declaration
    :param int uc: Uniqueness counter of ISCC-ID.
    :return: ISCC object with an ISCC-ID
    :rtype: dict
    """
    return gen_iscc_id_v0(iscc_code, chain_id, wallet, uc)


def gen_iscc_id_v0(iscc_code, chain_id, wallet, uc=0):
    # type: (str, int, str, Optional[int]) -> dict
    """
    Generate an ISCC-ID from an ISCC-CODE with uniqueness counter 'uc' with
    algorithm v0.

    :param str iscc_code: The ISCC-CODE from which to mint the ISCC-ID.
    :param int chain_id: Chain-ID of blockchain from which the ISCC-ID is minted.
    :param str wallet: The wallet address that signes the ISCC declaration
    :param int uc: Uniqueness counter of ISCC-ID.
    :return: ISCC object with an ISCC-ID
    :rtype: dict
    """
    if chain_id not in list(ic.ST_ID):
        raise AssertionError("Unknown Chain-ID {}".format(chain_id))
    iscc_id_digest = soft_hash_iscc_id_v0(iscc_code, wallet, uc)
    iscc_id_len = len(iscc_id_digest) * 8
    iscc_id = ic.encode_component(
        mtype=ic.MT.ID,
        stype=chain_id,
        version=ic.VS.V0,
        bit_length=iscc_id_len,
        digest=iscc_id_digest,
    )
    iscc = "ISCC:" + iscc_id
    return dict(iscc=iscc)


def soft_hash_iscc_id_v0(iscc_code, wallet, uc=0):
    # type: (str, str, int) -> bytes
    """
    Calculate ISCC-ID hash digest from ISCC-CODE with algorithm v0.

    Accepts an ISCC-CODE or any sequence of ISCC-UNITs.

    :param str iscc_code: ISCC-CODE
    :param str wallet: The wallet address that signes the ISCC declaration
    :param int uc: Uniqueness counter for ISCC-ID.
    :return: Digest for ISCC-ID without header but including uniqueness counter.
    :rtype: bytes
    """
    components = ic.iscc_decompose(iscc_code)
    decoded = [ic.decode_base32(c) for c in components]
    unpacked = [ic.decode_header(d) for d in decoded]

    digests = []

    if len(unpacked) == 1 and unpacked[0][0] == ic.MT.INSTANCE:
        # Special case if iscc_code is a singular Instance-Code
        digests.append(decoded[0][:1] + unpacked[0][-1][:7])
    else:
        for dec, unp in zip(decoded, unpacked):
            mt = unp[0]
            if mt == ic.MT.INSTANCE:
                continue
            if mt == ic.MT.ID:
                raise ValueError("Cannot create ISCC-ID from ISCC-ID")
            # first byte of header + first 7 bytes of body
            digests.append(dec[:1] + unp[-1][:7])

    iscc_id_digest = ic.alg_simhash(digests)

    # XOR with sha2-256 of wallet
    wallet_hash_digest = sha256(wallet.encode("ascii")).digest()[:8]
    iscc_id_xor_digest = bytes(a ^ b for (a, b) in zip(iscc_id_digest, wallet_hash_digest))

    if uc:
        iscc_id_xor_digest += uvarint.encode(uc)
    return iscc_id_xor_digest


def iscc_id_incr(iscc_id):
    # type: (str) -> str
    """
    Increment uniqueness counter of an ISCC-ID with latest standard algorithm.

    :param str iscc_id: Base32-encoded ISCC-ID.
    :return: Base32-encoded ISCC-ID with counter incremented by one.
    :rtype: str
    """
    return iscc_id_incr_v0(iscc_id)


def iscc_id_incr_v0(iscc_id):
    # type: (str) -> str
    """
    Increment uniqueness counter of an ISCC-ID with algorithm v0.

    :param str iscc_id: Base32-encoded ISCC-ID.
    :return: Base32-encoded ISCC-ID with counter incremented by one (without "ISCC:" prefix).
    :rtype: str
    """
    clean = ic.iscc_clean(iscc_id)
    code_digest = ic.decode_base32(clean)
    mt, st, vs, ln, data = ic.decode_header(code_digest)
    if mt != ic.MT.ID:
        raise AssertionError(f"MainType {mt} is not ISCC-ID")
    if st not in list(ic.ST_ID):
        raise AssertionError(f"Unsupported chain-id {st}")
    if vs != ic.VS.V0:
        raise AssertionError(f"Version {vs} is not v0")
    if len(data) == 8:
        data += uvarint.encode(1)
    else:
        counter = uvarint.decode(data[8:])
        suffix = uvarint.encode(counter.integer + 1)
        data = data[:8] + suffix

    iscc_id_len = len(data) * 8
    iscc_id = ic.encode_component(
        mtype=mt,
        stype=st,
        version=vs,
        bit_length=iscc_id_len,
        digest=data,
    )

    return iscc_id


def alg_simhash_from_iscc_id(iscc_id, wallet):
    # type: (str, str) -> str
    """
    Extract similarity preserving hex-encoded hash digest from ISCC-ID

    We need to un-xor the ISCC-ID hash digest with the wallet address hash to obtain the similarity
    preserving bytestring.
    """
    wallet_hash_digest = sha256(wallet.encode("ascii")).digest()[:8]
    cleaned = ic.iscc_clean(iscc_id)
    iscc_tuple = ic.iscc_decode(cleaned)
    iscc_id_xor_digest = iscc_tuple[4][:8]
    iscc_id_digest = bytes(a ^ b for (a, b) in zip(iscc_id_xor_digest, wallet_hash_digest))
    return iscc_id_digest.hex()



# -*- coding: utf-8 -*-
"""An application that claims ISCC conformance MUST pass all core functions from the
ISCC conformance test suite. The test suite is available as JSON data on
[Github](https://raw.githubusercontent.com/iscc/iscc-core/master/iscc_core/data.json).
Test data is structured as follows:

```json
{
  "<function_name>": {
    "<test_name>": {
      "inputs": ["<value1>", "<value2>"],
      "outputs": ["value1>", "<value2>"]
      }
   }
}
```

Inputs that are expected to be `raw bytes or byte-streams` are embedded as HEX encoded strings
in JSON and prefixed with `stream:` or `bytes` to support automated decoding during
implementation testing.

!!! example
    Byte-stream outputs in JSON test data:
    ```json
    "gen_data_code_v0": {
      "test_0000_two_bytes_64": {
        "inputs": [
          "stream:ff00",
          64
        ],
        "outputs": {
          "iscc": "GAAXL2XYM5BQIAZ3"
        }
      },
      ...
    ```
"""
from loguru import logger as log
import pathlib
import json
import io
from typing import Generator, Tuple, Callable, Any, List
import iscc_core as ic


HERE = pathlib.Path(__file__).parent.absolute()
TEST_DATA = HERE / "data.json"


def conformance_testdata():
    # type: () -> Generator[Tuple[str, Callable, List[Any], List[Any]]]
    """
    Yield tuples of test data.

    :return: Tuple with testdata (test_name, func_obj, inputs, outputs)
    :rtype: Generator[Tuple[str, Callable, List[Any], List[Any]]]
    """
    with open(TEST_DATA, "rb") as stream:
        data = json.load(stream)
    for func_name, tests in data.items():
        func_obj = getattr(ic, func_name)
        for test_name, test_values in tests.items():
            # Convert stream and bytes test values
            ntv = []
            for tv in test_values["inputs"]:
                if isinstance(tv, str) and tv.startswith("stream:"):
                    ntv.append(io.BytesIO(bytes.fromhex(tv.lstrip("stream:"))))
                else:
                    ntv.append(tv)
            test_values["inputs"] = ntv

            yield test_name, func_obj, test_values["inputs"], test_values["outputs"]


def conformance_selftest():  # pragma: no cover
    # type: () -> bool
    """
    Run conformance tests.

    :return: whether all tests passed
    :rtype: bool
    """
    passed = True
    for test_name, func, inputs, outputs in conformance_testdata():
        try:
            if func(*inputs) != outputs:
                raise AssertionError(f"FAILED: {func.__name__}.{test_name}")
            log.info(f"PASSED: {func.__name__}.{test_name}")
        except Exception as e:
            log.error(f"FAILED: {func.__name__}.{test_name} with {e}")
            passed = False
    return passed
